import tensorflow as tf
import tf_keras as tfk
from tensorflow_probability import distributions as tfd
from abismal.likelihood.location_scale import LocationScale

class NormalLikelihood(LocationScale):
    def _likelihood(self, iobs, sigiobs):
        return tfd.Normal(iobs, sigiobs)




title = "Ray options"
description = "Ray is used to parallelize CrystFEL .stream file loading."

args_and_kwargs = (
    (
        (
            "--ray-log-level",
        ),{
            "help": "Ray log level which defaults to 'ERROR'.",
            "default": 'ERROR',
            "type": str,
        }
    ),
)



import numpy as np
import reciprocalspaceship as rs
import tensorflow as tf
from tensorflow_probability import distributions as tfd
from tensorflow_probability import layers  as tfl
from tensorflow_probability import util as tfu
from tensorflow_probability import bijectors as tfb
import tf_keras as tfk

class PosteriorBase(tfk.models.Model):
    def __init__(self, rac, epsilon=1e-12, kl_weight=1., **kwargs):
        """
        rac : ReciprocalASUCollection
        """
        super().__init__(**kwargs)
        self.epsilon = epsilon
        self.kl_weight = kl_weight
        self.rac = rac
        self.seen = self.add_weight(
            shape=self.rac.asu_size,
            initializer='zeros',
            dtype='bool',
            trainable=False,
            name="hkl_tracker",
        )

    def register_seen(self, asu_id, hkl):
        unique,_ = tf.unique(tf.reshape(self.rac._miller_ids(asu_id, hkl), [-1]))
        unique = unique[unique!=-1]
        seen_batch = tf.scatter_nd(
            unique[:,None], 
            tf.ones_like(unique, dtype='bool'), 
            shape=[self.rac.asu_size]
        )
        self.seen.assign(self.seen | seen_batch)

    def flat_distribution(self):
        raise NotImplementedError("Subclasses must implement a flat_distribution method")

    def flat_prior(self):
        raise NotImplementedError("Subclasses must implement a flat_prior method")

    def to_datasets(self, seen=True):
        """
        Parameters
        ----------
        seen : bool (optional)
            Only include reflections seen during training. Defaults to True. 
        """
        raise NotImplementedError("Subclasses must implement a to_datasets method")

    def mean(self, asu_id, hkl):
        q = self.flat_distribution
        mean = self.rac.gather(q.mean(), asu_id, hkl)
        return mean

    def stddev(self, asu_id, hkl):
        q = self.flat_distribution
        stddev = self.rac.gather(q.stddev(), asu_id, hkl)
        return stddev

    def register_kl(self, samples=None, asu_id=None, hkl=None, training=None):
        """
        this method will always try to use an analytical kl divergence. failing
        that, the samples will be used
        """
        if training:
            q,p = self.flat_distribution(), self.flat_prior()
            try:
                kl_div = q.kl_divergence(p)
            except NotImplementedError:
                kl_div = q.log_prob(samples) - p.log_prob(samples)
            kl_div = tf.reduce_mean(kl_div)
            self.add_metric(kl_div, name='KL')
            self.add_loss(self.kl_weight * kl_div)

    def get_flat_fsigf(self):
        msg = """
        Subclasses must implement get_flat_isigi and/or get_flat_fsigf
        """
        raise NotImplementedError(msg)

    def get_flat_isigi(self):
        msg = """
        Subclasses must implement get_flat_isigi and/or get_flat_fsigf
        """
        raise NotImplementedError(msg)

    def to_datasets(self, seen=True):
        h,k,l = self.rac.Hunique.numpy().T
        q = self.flat_distribution()
        data = {

            'H' : rs.DataSeries(h, dtype='H'),
            'K' : rs.DataSeries(k, dtype='H'),
            'L' : rs.DataSeries(l, dtype='H'),
        }

        try:
            I,SIGI = self.get_flat_isigi()
            has_isigi = True
            data.update({
                'I' : rs.DataSeries(I, dtype='J'),
                'SIGI' : rs.DataSeries(SIGI, dtype='Q'),
            })
        except NotImplementedError:
            has_isigi = False

        try:
            F,SIGF = self.get_flat_fsigf()
            has_fsigf = True
            data.update({
                'F' : rs.DataSeries(F, dtype='F'),
                'SIGF' : rs.DataSeries(SIGF, dtype='Q'),
            })
        except NotImplementedError:
            has_fsigf = False

        if not (has_fsigf or has_isigi):
            raise NotImplementedError


        asu_id = self.rac.asu_id
        for i,rasu in enumerate(self.rac):
            ds = rs.DataSet(
                data,
                merged=True,
                cell=rasu.cell,
                spacegroup=rasu.spacegroup,
            )
            idx = self.rac.asu_id.numpy() == i
            if seen:
                idx = idx & self.seen.numpy()

            out = ds[idx]
            out = out.set_index(['H', 'K', 'L'])
            if rasu.anomalous:
                out = out.unstack_anomalous()
                keys = []
                if has_fsigf:
                    keys += [
                        'F(+)',
                        'SIGF(+)',
                        'F(-)',
                        'SIGF(-)',
                    ]
                if has_isigi:
                    keys += [
                        'I(+)',
                        'SIGI(+)',
                        'I(-)',
                        'SIGI(-)',
                    ]
                out = out[keys]
            yield out

class StructureFactorPosteriorBase(PosteriorBase):
    parameterization = 'structure_factor'

    def get_flat_fsigf(self):
        q = self.flat_distribution()
        F = q.mean()      
        SIGF = q.stddev()
        return F, SIGF

    def get_flat_isigi(self):
        """
        This method is approximate, but the intensity is exact. It is calculated 
        based on the definition of variance. 
        The uncertainties are based on 1st order uncertainty propagation and are
        therefore less accurate. 
        Subclasses may choose to implement analytical expressions where available.
        """
        F,SIGF = self.get_flat_fsigf()
        I = F*F + SIGF*SIGF
        SIGI = np.abs(2*F*SIGF)
        return I,SIGI

class IntensityPosteriorBase(PosteriorBase):
    parameterization = 'intensity'

    def get_flat_isigi(self):
        q = self.flat_distribution()
        I = q.mean()      
        SIGI = q.stddev()
        return I, SIGI

    def get_flat_fsigf(self):
        msg = """
        Subclasses may implement get_flat_fsigf if the parameterization allows
        if the moments of the square root of the intensity distribution are known.
        """
        raise NotImplementedError(msg)




import tensorflow as tf
import tf_keras as tfk
from tensorflow_probability import distributions as tfd
from abismal.likelihood.location_scale import LocationScale

class StudentTLikelihood(LocationScale):
    def __init__(self, degrees_of_freedom):
        super().__init__()
        self.degrees_of_freedom = degrees_of_freedom

    def _likelihood(self, iobs, sigiobs):
        return  tfd.StudentT(self.degrees_of_freedom, iobs, sigiobs)




title = "Surrogate posteriors"
description = "Arguments affecting the parameterization and initialization of surrogate posteriors"

args_and_kwargs = (
    (
        (
            "--init-scale",
        ),{
            "help": "The surrogate posteriors will be initialized with standard deviations equal to init_scale * mean(prior). The default is 0.01. ",
            "default": 0.01,
            "type": float,
        }
    ),

    (
        (
            "--anomalous",
        ),{
            "help": "Keep the two halves of reciprocal space separate during merging.",
            "action": 'store_true',
        }
    ),

    (
        (
            "--intensity-posterior",
        ),{
            "help": "Use a gamma-distributed posterior over intensities instead of the"
                    "default FoldedNormal over structure factors.",
            "action": 'store_true',
        }
    ),

    (
        (
            "--cell",
        ),{
            "help": "Use the specified unit cell.",
            "nargs": 6,
            "type": float,
            "default" : None,
        }
    ),

    (
        (
            "--space-group",
        ),{
            "help": "Use the specified space group instead of P1.",
            "type": str,
            "default" : None,
        }
    ),

    (
        (
            "-d",
            "--dmin",
        ),{
            "help": "The resolution at which to truncate the data",
            "required": True,
            "type": float,
        }
    ),
)



import numpy as np
import reciprocalspaceship as rs
import tensorflow as tf
from tensorflow_probability import distributions as tfd
from tensorflow_probability import util as tfu
from tensorflow_probability import bijectors as tfb
import tf_keras as tfk
from abismal.distributions import FoldedNormal as FoldedNormal
from abismal.surrogate_posterior import StructureFactorPosteriorBase
from abismal.surrogate_posterior.structure_factor.wilson import WilsonPrior


class FoldedNormalPosterior(StructureFactorPosteriorBase):
    def __init__(self, rac, prior=None, scale_factor=1e-2, epsilon=1e-12, kl_weight=1., **kwargs):
        super().__init__(rac, epsilon=epsilon, kl_weight=kl_weight, **kwargs)
        if prior is None:
            self._flat_prior = WilsonPrior(
                self.rac.centric,
                self.rac.epsilon,
            )
        else:
            self._flat_prior = prior

        self.low = self.epsilon * tf.cast(~self.rac.centric, dtype='float32')
        p = self.flat_prior()
        loc_init = p.mean()
        self.loc = tfu.TransformedVariable(
            loc_init,
            tfb.Exp(),
        )

        self.scale = tfu.TransformedVariable(
            scale_factor * loc_init,
            tfb.Chain([
                tfb.Shift(epsilon), 
                tfb.Exp(),
            ]),
        )

    def flat_prior(self):
        return self._flat_prior

    def flat_distribution(self):
        f = FoldedNormal(
            self.loc, 
            self.scale, 
        )
        q = tfd.TransformedDistribution(
            f, 
            tfb.Shift(self.low),
        )
        return q




title = "TensorFlow options"
description = "Options dictating how and what resources will be used by TF"

args_and_kwargs = (
    (
        (
            "--gpu-id",
        ),{
            "help": "The numerical id of the GPU you wish to use.",
            "default": 0,
            "type": int,
        }
    ),

    (
        (
            "--tf-log-level",
        ),{
            "help": "The tf log level is an int with default 3.",
            "default": 3,
            "type": int,
        }
    ),

    (
        (
            "--run-eagerly",
        ),{
            "help": "Run in eager mode.",
            "action" : "store_true",
        }
    ),
)



import numpy as np
import tensorflow as tf
import tf_keras as tfk
from tensorflow_probability import stats as tfs


class Standardize(tfk.layers.Layer):
    def __init__(self, center=True, decay=0.999, epsilon=1e-3):
        super().__init__()
        self.decay = decay
        self.center = center
        self.epsilon = epsilon

    def build(self, shape):
        d = shape[-1]
        self._mean = self.add_weight(
            shape=d,
            initializer='zeros',
            dtype=tf.float32,
            trainable=False,
            name='mean',
        )
        self._var = self.add_weight(
            shape=d,
            initializer='zeros',
            dtype=tf.float32,
            trainable=False,
            name='variance',
        )
        self.count = self.add_weight(
            shape = (),
            initializer = 'zeros',
            dtype=tf.int32,
            trainable=False,
            name='zero_debias_count',
        )

    def _debiased_mean_variance(self):
        mean,var = tfs.moving_mean_variance_zero_debiased(
            self._mean,
            self._var,
            self.count,
            decay=self.decay,
        )
        return mean,var

    @property
    def mean(self):
        mean,_ = self._debiased_mean_variance()
        return mean

    @property
    def var(self):
        _,var = self._debiased_mean_variance()
        return var

    @property
    def std(self):
        s = tf.sqrt(self.var)
        return tf.clip_by_value(s, self.epsilon, np.inf)

    def update(self, x):
        tfs.assign_moving_mean_variance(
            x,
            self._mean,
            self._var,
            zero_debias_count=self.count,
            decay=self.decay,
            axis=-2,
        )

    def standardize(self, data):
        mean,var = self._debiased_mean_variance()
        std = tf.clip_by_value(tf.sqrt(var), self.epsilon, np.inf)
        if self.center:
            return (data - mean) / std
        return data / std

    def call(self, data, training=True):
        if training:
            self.update(data)
        return self.standardize(data)






def set_log_level(level):
    from os import environ
    environ['TF_CPP_MIN_LOG_LEVEL'] =  str(int(level))

# Handle GPU selection
def set_gpu(gpu_id):
    import tensorflow as tf
    gpus = tf.config.experimental.list_physical_devices('GPU')
    for g in gpus:
        tf.config.experimental.set_memory_growth(g, True)
    if gpus:
        try:
            if gpu_id is None:
                tf.config.experimental.set_visible_devices([], 'GPU')
            else:
                tf.config.experimental.set_visible_devices(gpus[gpu_id], 'GPU')
        except RuntimeError as e:
            # Visible devices must be set at program startup
            print(e)




"""
ABISMAL - merge serial diffraction data using neural networks and variational inference.
"""

import abismal.command_line.parser.io as io
import abismal.command_line.parser.architecture as architecture
import abismal.command_line.parser.likelihood as likelihood
import abismal.command_line.parser.optimizer as optimizer
import abismal.command_line.parser.phenix as phenix
import abismal.command_line.parser.priors as priors
import abismal.command_line.parser.ray as ray
import abismal.command_line.parser.surrogate_posterior as surrogate_posterior
import abismal.command_line.parser.tf as tf
import abismal.command_line.parser.training as training

groups = [
    architecture,
    io,
    likelihood,
    optimizer,
    phenix,
    priors,
    ray,
    surrogate_posterior,
    tf,
    training,
]

from argparse import ArgumentParser
parser = ArgumentParser(description=__doc__)
for group in groups:
    g = parser.add_argument_group(group.title, group.description)
    for args,kwargs in group.args_and_kwargs:
        g.add_argument(*args, **kwargs)

all = [
    parser,
]



import tensorflow as tf
from tensorflow_probability import distributions as tfd
import numpy as np


n = 100
conc,rate = np.random.random((2, n))
q = tfd.Gamma(conc, rate)


def moments_sqrt(q):
    alpha = q.concentration
    beta  = q.rate
    m = alpha
    omega = alpha / beta
    mean = tf.math.exp(
        tf.math.lgamma(m + 0.5) - tf.math.lgamma(m) + 0.5 * tf.math.log(omega) - 0.5 * tf.math.log(m)
    )
    var = omega - mean*mean
    std = tf.math.sqrt(var)
    return mean,std


s = 10_000
f,sigf = moments_sqrt(q)

z = q.sample(s)
z = tf.sqrt(z)
f_mc = tf.reduce_mean(z, axis=0)
sigf_mc = tf.math.reduce_std(z, axis=0)
from IPython import embed
embed(colors='linux')



import tensorflow as tf
import tf_keras as tfk
import gemmi

class Op(tfk.layers.Layer):
    def __init__(self, triplet):
        super().__init__()
        self.gemmi_op = gemmi.Op(triplet)
        self.rot = tf.convert_to_tensor(self.gemmi_op.rot, dtype='float32')
        self.den = tf.convert_to_tensor(self.gemmi_op.DEN, dtype='float32')
        self.identity = self.gemmi_op == 'x,y,z'

    def __str__(self):
        return f"Op({self.gemmi_op.triplet()})"

    def call(self, hkl):
        if self.identity:
            return hkl
        dtype = hkl.dtype
        hkl = tf.cast(hkl, tf.float32)
        hkl = tf.math.floordiv(tf.matmul(hkl, self.rot), self.den)
        hkl = tf.cast(hkl, dtype)
        return hkl




#/usr/bin/env cctbx.python
import numpy as np
import argparse
import reciprocalspaceship as rs
import pandas as pd
import gemmi
import tensorflow as tf


class DataLoader():
    data_labels = (
        "ASU",
        "HKL",
        "Resolution",
        "Wavelength",
        "Metadata",
        "I",
        "SigI",
    ) #To support laue add wavelength and resolution here
    def __init__(self, metadata_length):
        self.signature=(
            (
                tf.RaggedTensorSpec((None, None, 1), tf.int32, 1, tf.int32), #ASU ID
                tf.RaggedTensorSpec((None, None, 3), tf.int32, 1, tf.int32), #HKL 
                tf.RaggedTensorSpec((None, None, 1), tf.float32, 1, tf.int32), #dHKL
                tf.RaggedTensorSpec((None, None, 1), tf.float32, 1, tf.int32), #wavelength
                tf.RaggedTensorSpec((None, None, metadata_length), tf.float32, 1, tf.int32), #metadata
                tf.RaggedTensorSpec((None, None, 1), tf.float32, 1, tf.int32), #I
                tf.RaggedTensorSpec((None, None, 1), tf.float32, 1, tf.int32), #SigI
            ),(
                tf.RaggedTensorSpec((None, None, 1), tf.float32, 1, tf.int32), #I 
            )
        )




import tensorflow as tf
from tensorflow.python.ops import array_ops
from tensorflow_probability.python.internal import dtype_util
from tensorflow_probability.python.internal import special_math
from tensorflow_probability import distributions as tfd
from tensorflow_probability.python.internal import reparameterization
from tensorflow_probability.python.internal import prefer_static as ps
import numpy as np
from tensorflow_probability.python.bijectors import exp as exp_bijector
from tensorflow_probability.python.bijectors import absolute_value as abs_bijector
from tensorflow_probability.python.internal import parameter_properties
from tensorflow_probability.python.internal import samplers
from tensorflow_probability import math as tfm
from tensorflow_probability.python.internal import tensor_util
from tensorflow_probability import math as tfm
import math

def folded_normal_sample_gradients(z, loc, scale):
    alpha = (z + loc) / scale
    beta = (z - loc) / scale

    p_a = tfd.Normal(loc, scale).prob(z)   #N(z|loc,scale)
    p_b = tfd.Normal(-loc, scale).prob(z)  #N(z|-loc,scale)

    # This formula is dz = N(z|-loc,scale) + N(z|loc,scale)
    dz = p_a + p_b

    # This formula is dloc = N(z|-loc,scale) - N(z|loc,scale)
    dloc = p_b - p_a
    dloc = dloc / dz

    # This formula is -[b * N(z|-loc, scale) + a * N(z|loc, scale)]
    dscale = -alpha * p_b - beta * p_a
    dscale = dscale / dz
    return dloc, dscale

@tf.custom_gradient
def stateless_folded_normal(shape, loc, scale, seed):
    z = tf.random.stateless_normal(shape, seed, mean=loc, stddev=scale)
    z = tf.abs(z)
    def grad(upstream):
        dloc,dscale = folded_normal_sample_gradients(z, loc, scale)
        dloc = tf.reduce_sum(-upstream * dloc, axis=0)
        dscale = tf.reduce_sum(-upstream * dscale, axis=0)
        return None, dloc, dscale, None
    return z, grad

class FoldedNormal(tfd.Distribution):
    """The folded normal distribution."""
    _normal_crossover = 10.
    def __init__(self,
               loc,
               scale,
               validate_args=False,
               allow_nan_stats=True,
               name='FoldedNormal'):
        """Construct a folded normal distribution.
        Args:
          loc: Floating-point `Tensor`; the means of the underlying
            Normal distribution(s).
          scale: Floating-point `Tensor`; the stddevs of the underlying
            Normal distribution(s).
          validate_args: Python `bool`, default `False`. Whether to validate input
            with asserts. If `validate_args` is `False`, and the inputs are
            invalid, correct behavior is not guaranteed.
          allow_nan_stats: Python `bool`, default `True`. If `False`, raise an
            exception if a statistic (e.g. mean/mode/etc...) is undefined for any
            batch member If `True`, batch members with valid parameters leading to
            undefined statistics will return NaN for this statistic.
          name: The name to give Ops created by the initializer.
        """
        parameters = dict(locals())
        with tf.name_scope(name) as name:
          dtype = dtype_util.common_dtype([loc, scale], dtype_hint=tf.float32)
          self._loc = tensor_util.convert_nonref_to_tensor(
              loc, dtype=dtype, name='loc')
          self._scale = tensor_util.convert_nonref_to_tensor(
              scale, dtype=dtype, name='scale')
          super(FoldedNormal, self).__init__(
              dtype=dtype,
              reparameterization_type=reparameterization.FULLY_REPARAMETERIZED,
              validate_args=validate_args,
              allow_nan_stats=allow_nan_stats,
              parameters=parameters,
              name=name)

    def _batch_shape_tensor(self, loc, scale):
        return array_ops.shape(loc / scale)

    @classmethod
    def _parameter_properties(cls, dtype, num_classes=None):
        return dict(
            loc=parameter_properties.ParameterProperties(),
            scale=parameter_properties.ParameterProperties(
                default_constraining_bijector_fn=(
                    lambda: softplus_bijector.Softplus(low=dtype_util.eps(dtype))
                )
            ),
        )

    @property
    def loc(self):
        """Distribution parameter for the pre-transformed mean."""
        return self._loc

    @property
    def scale(self):
        """Distribution parameter for the pre-transformed standard deviation."""
        return self._scale

    def _event_shape_tensor(self):
        return tf.constant([], dtype=tf.int32)

    def _event_shape(self):
        return tf.TensorShape([])

    def _cdf(self, x):
        loc,scale = self.loc,self.scale
        x = tf.convert_to_tensor(x)
        a = (x + loc) / scale
        b = (x - loc) / scale
        ir2 = tf.constant(tf.math.reciprocal(tf.sqrt(2.)), dtype=x.dtype)
        return 0.5 * (tf.math.erf(ir2 * a) - tf.math.erf(ir2 * b))

    def _sample_n(self, n, seed=None):
        seed = samplers.sanitize_seed(seed)
        loc = tf.convert_to_tensor(self.loc)
        scale = tf.convert_to_tensor(self.scale)
        shape = ps.concat([[n], self._batch_shape_tensor(loc=loc, scale=scale)], axis=0)
        return stateless_folded_normal(shape, loc, scale, seed)

    def _log_prob(self,  value):
        loc,scale = self.loc,self.scale
        result = tfm.log_add_exp(
            tfd.Normal(loc, scale).log_prob(value), 
            tfd.Normal(-loc, scale).log_prob(value),
        )
        return result

    @staticmethod
    def _folded_normal_mean(loc, scale):
        u,s = loc, scale
        c = loc / scale
        mean = (
            s * tf.sqrt(2/math.pi) * tf.math.exp(-0.5 * c * c) + 
            u * tf.math.erf(c/math.sqrt(2))
        )
        return mean

    def _mean(self):
        u = self.loc
        s = self.scale
        c = u/s

        idx = tf.abs(c) >= self._normal_crossover
        s_safe = tf.where(idx, 1., s)
        u_safe = tf.where(idx, 1., u)

        return tf.where(
            idx,
            tf.abs(u),
            self._folded_normal_mean(u_safe, s_safe)
        )

    def _variance(self):
        u = self.loc
        s = self.scale
        c = u/s

        idx = tf.abs(c) >= 10.
        s_safe = tf.where(idx, 1., s)
        u_safe = tf.where(idx, 1., u)
        m = self._folded_normal_mean(u_safe, s_safe)

        return tf.where(
            idx, 
            s*s,
            u*u + s*s - m*m,
        )

    def moment(self, t):
        """ Use Scipy to calculate the t-moment of a folded normal """
        from scipy.stats import foldnorm,norm
        loc,scale = self.loc.numpy(),self.scale.numpy()
        c = loc / scale
        idx = np.abs(c) > self._normal_crossover
        c_safe = tf.where(idx, c, 0.)
        scale_safe = tf.where(idx, scale, 1.)

        result = np.where(
            idx,
            foldnorm.moment(t, loc/scale, scale=scale),
            norm.moment(t, loc=loc, scale=scale),
        )
        return result



import numpy as np
import tensorflow as tf
import reciprocalspaceship as rs
from abismal.distributions import TruncatedNormal,FoldedNormal
from tensorflow_probability import distributions as tfd
from tensorflow_probability import layers  as tfl
from tensorflow_probability import util as tfu
from tensorflow_probability import bijectors as tfb
from tensorflow_probability import stats as tfs
import tf_keras as tfk

from abismal.layers import *
from abismal.distributions import FoldedNormal

class LocationScaleLayer(tfk.layers.Layer):
    def __init__(
            self, 
            eps=1e-12,
            kernel_initializer='glorot_normal',
            scale_bijector=None,
            loc_bijector=None,
        ):
        super().__init__()
        self.dense = tfk.layers.Dense(2, kernel_initializer=kernel_initializer)
        self.eps = eps
        self.scale_bijector = scale_bijector
        if scale_bijector is None:
            self.scale_bijector = tfb.Chain([tfb.Shift(eps), tfb.Exp()])
        self.loc_bijector = loc_bijector

class FoldedNormalLayer(LocationScaleLayer):
    def call(self, data, training=None, mc_samples=None, **kwargs):
        theta = self.dense(data)
        loc, scale = tf.unstack(theta, axis=-1)

        if self.scale_bijector is not None:
            scale = self.scale_bijector(scale)
        if self.loc_bijector is not None:
            loc = self.loc_bijector(loc)

        q = FoldedNormal(loc, scale)
        return q

class DeltaDist():
    def __init__(self, loc):
        self.loc = loc

    def sample(self, *args, **kwargs):
        return self.loc[None,...]

class DeltaLayer(tfk.layers.Layer):
    def __init__(self, kernel_initializer='glorot_normal', loc_bijector=None):
        super().__init__()
        self.loc_bijector = loc_bijector
        self.dense = tfk.layers.Dense(1, kernel_initializer=kernel_initializer)

    def call(self, data, *args, **kwargs):
        loc = self.dense(data)
        if self.loc_bijector is not None:
            loc = self.loc_bijector(loc)
        loc = tf.squeeze(loc, axis=-1)
        return DeltaDist(loc)

class ImageScaler(tfk.layers.Layer):
    def __init__(
            self, 
            mlp_width=32, 
            mlp_depth=20, 
            hidden_units=None,
            activation="ReLU",
            kernel_initializer=None,
            scale_posterior=None,
            scale_prior=None,
            kl_weight=1.,
            eps=1e-12,
            num_image_samples=32,
            share_weights=True,
            seed=None,
            **kwargs, 
        ):
        """
        This function has a lot of overrides, but comes with sensible defaults built in. 

        Parameters
        ----------
        mlp_width : int (optional)
            Default 32 neurons
        mlp_depth : int (optional)
            Default 20 layers
        hidden_units : int (optional)
            This defaults to 2*mlp_width
        activation : str (optional)
            This is a Keras activation function with the default being "ReLU"
        kernel_initializer : str or keras Initializer
            This defaults to a carefully chosen strategy which ensures the output is not
            too large. Users may want to override it for something more standard. 
        scale_posterior : layer (optional)
            Override the scale_posterior which defaults to a FoldedNormal if kl_weight > 0.
        scale_prior : Distribution (optional)
            Override the scale_prior which defaults to an exponential distribution
        kl_weight : float (optional)
            This toggles the behavior of the scale_posterior. If scale_posterior isn't passed, and
            kl_weight is 0., a Delta distribution is used. Otherwise, the default surrogate
            posterior is a FoldedNormal. 
        eps : float (optional)
            A small constant for numerical stability defaults to 1e-12. 
        num_image_samples : int (optional)
            The number of reflections to sample in order to create the image representation vectors. 
            The default is 32 samples. 
        share_weights : bool (optional)
            Whether or not share neural network weights between the image model and the scale model. 
            The default is True. 
        seed : int (optional)
            An int or tf random seed for initialization. 
        """
        super().__init__(**kwargs)

        self.kl_weight = kl_weight
        self.num_image_samples = num_image_samples
        self.scale_prior = scale_prior
        self.scale_posterior = scale_posterior

        if hidden_units is None:
            hidden_units = 2 * image_mpl_width

        if kernel_initializer is None:
            if seed is None:
                seed = 1234 #This needs to be set to something otherwise all layers will
                #initialize the same
            kernel_initializer = tfk.initializers.VarianceScaling(
                scale=1./10./mlp_depth,
                mode='fan_avg', distribution='truncated_normal', seed=seed
            )

        input_image   = [
            tfk.layers.Dense(mlp_width, kernel_initializer=kernel_initializer),
        ]
        input_scale = [
            tfk.layers.Dense(mlp_width, kernel_initializer=kernel_initializer),
        ]
            
        self.input_image   = tfk.models.Sequential(input_image)
        self.input_scale   = tfk.models.Sequential(input_scale)

        self.pool = Average(axis=-2)

        self.image_network = tfk.models.Sequential([
                FeedForward(
                    hidden_units=hidden_units,
                    activation=activation,
                    kernel_initializer=kernel_initializer,
                ) for i in range(mlp_depth)])

        if share_weights:
            self.scale_network = self.image_network
        else:
            self.scale_network = tfk.models.Sequential([
                FeedForward(
                    hidden_units=hidden_units, 
                    kernel_initializer=kernel_initializer, 
                    activation=activation, 
                    ) for i in range(mlp_depth)
            ]) 

        # If kl_weight is nonzero, set the default 
        # posterior to FoldedNormal, prior to Exponential
        if kl_weight > 0.:
            if self.scale_prior is None:
                self.scale_prior = tfd.Exponential(1.)
            if self.scale_posterior is None:
                self.scale_posterior = FoldedNormalLayer(
                    kernel_initializer=kernel_initializer,
                    eps=eps,
                )
        # if kl_weight is zero, set the default posterior to a Delta function
        else:
            if self.scale_posterior is None:
                self.scale_posterior = DeltaLayer(
                    kernel_initializer=kernel_initializer,
                )

    @staticmethod
    def sample_ragged_dim(ragged, mc_samples):
        """
        Randomly subsample "length" entries from ragged with replacement.
        """
        l = ragged.row_lengths()
        n = tf.shape(ragged)[0]
        r = tf.random.uniform((n, mc_samples))
        idx2 = tf.round(r * tf.cast(l, 'float32')[:,None]- 0.5)
        idx2 = tf.cast(idx2, 'int32')
        idx1 = tf.range(n)[:,None] * tf.ones_like(idx2)
        idx = tf.stack((idx1, idx2), axis=-1)
        out = tf.gather_nd(ragged, idx)
        return out

    def call(self, inputs, mc_samples=32, training=None, **kwargs):
        (
            asu_id,
            hkl,
            resolution,
            wavelength,
            metadata,
            iobs,
            sigiobs,
        ) = inputs

        scale = metadata
        image = [iobs, sigiobs, metadata]

        image = tf.concat(image, axis=-1)
        image = ImageScaler.sample_ragged_dim(image, self.num_image_samples)

        image = self.input_image(image)
        scale = self.input_scale(scale)

        image = self.image_network(image)
        image = self.pool(image)
        scale = scale + image

        scale = self.scale_network(scale)
        q = self.scale_posterior(scale.flat_values, training=training)
        z = q.sample(mc_samples)

        if self.scale_prior is not None:
            try:
                kl_div = q.kl_divergence(self.scale_prior)
            except NotImplementedError:
                q_z = q.log_prob(z)
                p_z = self.scale_prior.log_prob(z)
                kl_div = q_z - p_z

            kl_div = tf.reduce_mean(kl_div)
            self.add_loss(self.kl_weight * kl_div)
            self.add_metric(kl_div, name='KL_Σ')
        z = tf.RaggedTensor.from_row_splits(tf.transpose(z), metadata.row_splits)

        return z



#/usr/bin/env cctbx.python
import numpy as np
import argparse
import reciprocalspaceship as rs
import pandas as pd
import gemmi
import tensorflow as tf
from .loader import DataLoader



class StillsLoader(DataLoader):
    """ DIALS stills loader """
    def __init__(self, expt_files, refl_files, spacegroup=None, cell=None, dmin=None, asu_id=0, include_eo=True):
        self.include_eo = include_eo
        if include_eo:
            super().__init__(5)
        else:
            super().__init__(2)
        from dxtbx.model.experiment_list import ExperimentListFactory  #defer defer defer defer
        self.expt_files = expt_files
        self.refl_files = refl_files
        self.dmin = dmin
        self.asu_id = asu_id

        elist_list = [ExperimentListFactory.from_json_file(expt_file, check_format=False) for expt_file in self.expt_files]
        self.cell,self.spacegroup = cell,spacegroup 
        if self.cell is None:
            self.cell = self.get_average_cell(elist_list)
        if self.spacegroup is None:
            self.spacegroup = self.get_space_group(elist_list)

        self.mean = None
        self.std  = None

    @staticmethod
    def get_average_cell(elist_list):
        """
        Parameters
        ----------
        elist_list : list
            A list of DIALS experiment lists 

        Returns
        -------
            cell : gemmi.UnitCell
        """
        cell = np.zeros(6)
        l = 0
        from tqdm import tqdm
        print("Determining unit cell ...")
        for elist in tqdm(elist_list):
            crystals = elist.crystals()
            cell += np.array([c.get_unit_cell().parameters() for c in crystals]).sum(0)
            l += len(crystals)
        cell = cell/l
        cell = gemmi.UnitCell(*cell)
        print(f"Average cell: {cell}")
        return cell

    @staticmethod
    def get_space_group(elist_list, check_consistent=False):
        """
        This method will raise a `ValueError` if all the space groups do not match. 

        Parameters
        ----------
        elist_list : list
            A list of DIALS experiment lists 

        Returns
        -------
            spacegroup :  gemmi.SpaceGroup
        """
        hms = elist_list[0].crystals()[0].get_space_group().type().universal_hermann_mauguin_symbol()
        if not check_consistent:
            return gemmi.SpaceGroup(hms)

        from tqdm import tqdm
        for elist in tqdm(elist_list):
            crystals = elist.crystals()
            _hms = np.array([c.get_space_group().type().universal_hermann_mauguin_symbol() for c in crystals])
            if not np.all(_hms == hms):
                cid = np.where(_hms != hms)[0][0]
                raise ValueError(f"Crystal {cid} has Universal Hermann Mauguin symbol {_hms[cid]} but {hms} was expected")
        return gemmi.SpaceGroup(hms)

    def _load_expt_refl(self, expt_and_refl):
        expt = expt_and_refl[0]
        refl = expt_and_refl[1]
        ragged = self.dials_to_ragged(expt, refl)
        ds = tf.data.Dataset.from_tensor_slices(ragged)
        return ds

    def get_dataset(self):
        """
        Convert dials monochromatic stills files to a tf.data.Dataset.
        """
        def data_gen():
            for expt,refl in list(zip(self.expt_files, self.refl_files)):
                data = self.dials_to_ragged(expt, refl)
                yield data
        return tf.data.Dataset.from_generator(data_gen, output_signature=self.signature).unbatch()

    def dials_to_ragged(self, expt_file, refl_file):
        """
        Convert dials monochromatic stills files to ragged tensors.
        """
        from dials.array_family import flex
        from dxtbx.model.experiment_list import ExperimentListFactory 
        table = flex.reflection_table().from_file(refl_file)
        elist = ExperimentListFactory.from_json_file(expt_file, check_format=False)

        idx = flex.size_t(np.array(table['id']))

        table.compute_d(elist)
        table["A_matrix"] = flex.mat3_double( [C.get_A() for C in elist.crystals()] ).select(idx)
        table["s0_vec"] = flex.vec3_double( [e.beam.get_s0() for e in elist] ).select(idx)
        table["wavelength"] = flex.double( [e.beam.get_wavelength() for e in elist] ).select(idx)

        h = table["miller_index"].as_vec3_double()
        Q = table["A_matrix"] * h
        Svec = Q + table["s0_vec"]
        Qobs = table['s0_vec'].norms() * Svec / Svec.norms() - table['s0_vec']

        hkl = np.array(h, dtype='int32')
        d = np.array(table['d'], dtype='float32')
        wavelength = np.array(table['wavelength'], dtype='float32')
        dQ = np.array(Q - Qobs, dtype='float32')
        xy = np.array(Svec, dtype='float32')[:,:2]
        batch = table['id'].as_numpy_array()
        idx = ~self.spacegroup.operations().systematic_absences(h)
        if self.dmin is not None:
            idx &= d >= self.dmin

        I  = np.array(table['intensity.sum.value'], dtype='float32')
        SigI  = np.array(np.sqrt(table['intensity.sum.variance']), dtype='float32')

        hkl = hkl[idx]
        d = d[idx, None]
        wavelength = wavelength[idx, None]
        dQ = dQ[idx]
        xy = xy[idx]
        batch = batch[idx]
        I = I[idx, None]
        SigI = SigI[idx, None]
        if self.include_eo:
            metadata = np.concatenate((xy, dQ), axis=-1)
        else:
            metadata = xy

        if self.mean is None:
            self.mean = (metadata.mean(0), I.mean(0), SigI.mean(0))
            self.std  = (metadata.std(0), I.std(0), SigI.std(0))

        #TODO: why is this still necessary even with Welford standardization???
        #metadata = (metadata - self.mean[0]) / self.std[0]
        #I = I / self.std[1]
        #SigI = SigI / self.std[1]

        hkl = tf.RaggedTensor.from_value_rowids(hkl, batch)
        d = tf.RaggedTensor.from_value_rowids(d, batch)
        wavelength = tf.RaggedTensor.from_value_rowids(wavelength, batch)
        metadata = tf.RaggedTensor.from_value_rowids(
            metadata,
            batch,
        )
        I = tf.RaggedTensor.from_value_rowids(I, batch)
        SigI = tf.RaggedTensor.from_value_rowids(SigI, batch)
        asu = tf.ones_like(I, dtype='int32') * self.asu_id

        data = ((asu, hkl, d, wavelength, metadata, I, SigI), (I,))

        return data

if __name__=='__main__':
    stream_file = '/mnt/raid/data/xtal/20210615_Neutze_collab/indexing_dark_before_merging_intensities.stream'
    test = np.random.random((1_000, 6))
    mean, variance = StreamLoader.online_mean_variance(test)
    assert np.all(np.isclose(mean, test.mean(0)))
    assert np.all(np.isclose(variance, np.square(test.std(0))))

    for crystal in StreamLoader.to_blocks(stream_file, 'crystal'):
        break

    assert len(crystal) > 0

    cell = StreamLoader.get_average_cell(stream_file)

    geo = StreamLoader.to_blocks(stream_file, 'geometry')
    print(''.join(next(geo)))

    print(StreamLoader.get_wavelength(stream_file))



import tensorflow as tf
import tf_keras as tfk
from tensorflow_probability import distributions as tfd
from tensorflow_probability import bijectors as tfb
from tensorflow_probability import util as tfu
from abismal.distributions.folded_normal import FoldedNormal

def make_fn_posterior(loc, scale, epsilon=1e-12):
    loc = tf.convert_to_tensor(loc)
    scale = tf.convert_to_tensor(scale)

    loc = tf.Variable(loc)
    scale = tfu.TransformedVariable(
        scale,
        tfb.Chain([
            tfb.Shift(epsilon),
            tfb.Exp(),
        ]),
    )
    return FoldedNormal(loc, scale)

class AdaptiveLikelihoodBase(tfk.layers.Layer):
    """
    A Bayesian additive error model inspired by EV11 wherein
    sigi = a * sqrt(sigiobs**2. + b * iobs + c * iobs**2.)
    a ~ HalfNormal(1.)
    """
    def __init__(self, prior=None, kl_weight=1., epsilon=1e-6, scale_factor=1e-2):
        super().__init__()
        self.kl_weight = kl_weight
        self.prior = prior
        if self.prior is None:
            self.prior = tfd.HalfNormal(1.)

        self.qa = make_fn_posterior(
            1.,
            scale_factor * self.prior.stddev(),
            epsilon=epsilon,
        )
        self.qb = make_fn_posterior(
            epsilon,
            scale_factor * self.prior.stddev(),
            epsilon=epsilon,
        )
        self.qc = make_fn_posterior(
            epsilon,
            scale_factor * self.prior.stddev(),
            epsilon=epsilon,
        )

    def _distribution(self, iobs, sigiobs):
        raise NotImplementedError('Extensions of this class must implement _distribution')

    def call(self, ipred, iobs, sigiobs):
        mc_samples = tf.shape(ipred)[-1]
        a = self.qa.sample(mc_samples)
        b = self.qb.sample(mc_samples)
        c = self.qc.sample(mc_samples)

        #var = sigiobs * sigiobs + b[...,:] * tf.abs(iobs) + c[...,:] * tf.square(iobs)
        var = sigiobs * sigiobs + b[...,:] * tf.abs(ipred) + c[...,:] * tf.square(ipred)
        sigi = a[...,:] * tf.sqrt(var)
        ll = self._distribution(iobs, sigi).log_prob(ipred)

        q = (
            self.qa.log_prob(a) +
            self.qb.log_prob(b) +
            self.qc.log_prob(c)
        )
        p = (
            self.prior.log_prob(a) +
            self.prior.log_prob(b) +
            self.prior.log_prob(c)
        )
        kl_div = tf.reduce_mean(q - p)
        self.add_metric(kl_div, "KL_LL")
        self.add_loss(self.kl_weight * kl_div)
        self.add_metric(self.qa.mean(), 'sdfac')
        self.add_metric(self.qb.mean(), 'sdadd')
        self.add_metric(self.qc.mean(), 'sdb')

        return ll

class AdaptiveNormalLikelihood(AdaptiveLikelihoodBase):
    def _distribution(self, iobs, sigiobs):
        return tfd.Normal(iobs, sigiobs)

class AdaptiveTLikelihood(AdaptiveLikelihoodBase):
    def __init__(self, dof, prior=None, kl_weight=1., epsilon=1e-6, scale_factor=1e-2):
        super().__init__(prior, kl_weight, epsilon, scale_factor)
        self.dof = dof

    def _distribution(self, iobs, sigiobs):
        return tfd.StudentT(self.dof, iobs, sigiobs)




import numpy as np
import reciprocalspaceship as rs
import tensorflow as tf
from tensorflow_probability import distributions as tfd
from tensorflow_probability import layers  as tfl
from tensorflow_probability import util as tfu
from tensorflow_probability import bijectors as tfb
import tf_keras as tfk
from IPython import embed

from abismal.layers import *
from abismal.distributions import FoldedNormal

class GraphImageScaler(tfk.layers.Layer):
    def __init__(self, 
        mlp_width, 
        mlp_depth, 
        dropout=0.0, 
        hidden_units=None,
        layer_norm=False,
        activation="ReLU",
        kernel_initializer='glorot_normal',
        stop_f_grad=True,
        num_heads=8,
        eps=1e-12,
        kl_weight=1.,
        scale_posterior=None,
        **kwargs, 
        ):
        super().__init__(**kwargs)


        if hidden_units is None:
            hidden_units = 2 * image_mpl_width

        self.input_image   = tfk.layers.Dense(mlp_width, kernel_initializer=kernel_initializer)
        self.input_scale  = tfk.layers.Dense(mlp_width, kernel_initializer=kernel_initializer)

        self.image_network = TransformerAverage(
            mlp_depth,
            num_heads=num_heads, key_dim=hidden_units,
            dropout=dropout,
            hidden_units=hidden_units,
            activation=activation,
            kernel_initializer=kernel_initializer,
            normalize=layer_norm,
        )

        self.scale_network = tfk.models.Sequential([
                FeedForward(
                    hidden_units=hidden_units, 
                    normalize=layer_norm, 
                    kernel_initializer=kernel_initializer, 
                    activation=activation, 
                    ) for i in range(mlp_depth)
            ] + [
                tfk.layers.Dense(1, kernel_initializer=kernel_initializer),
        ])

        if scale_posterior is None:
            prior = FoldedNormal(0., np.sqrt(np.pi/2).astype('float32'))
            self.scale_posterior = FoldedNormalLayer(
                kl_weight=kl_weight,
                prior=prior,
                kernel_initializer=kernel_initializer,
                eps=eps,
            )
        else:
            self.scale_posterior = scale_posterior

        self.stop_f_grad = stop_f_grad

    def call(self, inputs, mc_samples=1, training=None, **kwargs):
        metadata, iobs, sigiobs, imodel = inputs
        
        if self.stop_f_grad:
            imodel = tf.stop_gradient(imodel)

        scale = metadata
        image = tf.concat((metadata, iobs, sigiobs, imodel), axis=-1)

        image = self.input_image(image)
        scale = self.input_scale(scale)

        image = self.image_network(image)
        scale = scale + image

        scale = self.scale_network(scale)
        q = self.scale_posterior(scale.flat_values, training=training)
        z = q.sample(mc_samples)
        z = tf.RaggedTensor.from_row_splits(tf.transpose(z), metadata.row_splits)
        return z




import numpy as np
import reciprocalspaceship as rs
import tensorflow as tf
import tf_keras as tfk
from tensorflow_probability import distributions as tfd
from tensorflow_probability import layers  as tfl
from tensorflow_probability import util as tfu
from tensorflow_probability import bijectors as tfb


class FeedForward(tfk.layers.Layer):
    """
    This is a ResNet version 2 style layer
    """
    def __init__(self, 
        dropout=None, 
        hidden_units=None, 
        activation='ReLU',
        kernel_initializer='glorot_normal', 
        normalize=False, 
        **kwargs
        ):
        """
        This is a ResNet version 2 style nonlinearity. It implements the following

        ```
        out = dropout(linear(activation(hidden_linear(activation(layer_norm(in)))))) + in
        ```
        Where dropout and layer normalization are optional. 

        Parameters
        ----------
        dropout : float (optional)
            Apply dropout with this rate. Dropout occurs after the second linear layer. By default
            dropout is not used.
        hidden_units : int (optional)
            The size of the hidden layer. By default this will be 2 times the size of the input.
        activation : string or callable (optional)
            Either a string name of a keras activation or a callable function. The default is 'ReLU'.
        kernel_initializer : string or callable (optional)
            Either a string a keras intializer style function. The default is 'glorot_normal'. 
        normalize : bool (optional)
            Optionally apply layer normalization to the input. 
        """
        super().__init__()
        self.hidden_units = hidden_units
        self.kernel_initializer = kernel_initializer

        if dropout is not None:
            self.dropout = tfk.layers.Dropout(dropout)
        else:
            self.dropout = None

        if normalize:
            self.normalize = tfk.layers.LayerNormalization(axis=-1)
        else:
            self.normalize = None

        self.activation = tfk.activations.get(activation)

    def build(self, shape, **kwargs):
        self.units = shape[-1]
        if self.hidden_units is None:
            self.hidden_units = 2 * self.units

        self.ff1 = tfk.layers.Dense(self.hidden_units, kernel_initializer=self.kernel_initializer, **kwargs)
        self.ff2 = tfk.layers.Dense(self.units, kernel_initializer=self.kernel_initializer, **kwargs)

    def call(self, X, **kwargs):
        out = X

        if self.normalize is not None:
            out = self.normalize(out)

        out = self.activation(out)
        out = self.ff1(out)
        out = self.activation(out)
        out = self.ff2(out)

        if self.dropout is not None:
            out = self.dropout(out)

        out = out + X
        return out




title = "Architecture"
description = "Arguments affecting the model architecture and dimensions"

args_and_kwargs=(
    (
        (
            "--d-model",
        ),{
        "help": "The number of channels in the model with default 32.",
        "default" : 32,
        "type": int,
        }
    ),

    (
        (
            "--layers",
        ),{
            "help": "The number of feedfoward layers with default 20.",
            "default": 20,
            "type": int,
        }
    ),

    (
        (
            "--activation",
        ),{
            "help": "The name of the activation function used in the scale model. The default is 'relu'",
            "default": "relu",
            "type": str,
        }
    ),

    (
        (
            "--epsilon",
        ),{
            "help": "A small constant for numerical stability.",
            "default": 1e-12,
            "type": float,
        }
    ),
)



# Version number for reciprocalspaceship
def getVersionNumber():
    import pkg_resources

    version = pkg_resources.require("abismal")[0].version
    return version


__version__ = getVersionNumber()

from os import environ
environ['TF_CPP_MIN_LOG_LEVEL'] = environ.get("TF_CPP_MIN_LOG_LEVEL", "3")

from abismal import callbacks,distributions,io,layers,likelihood,merging,optimizers,scaling,surrogate_posterior,symmetry



import numpy as np
import reciprocalspaceship as rs
import tensorflow as tf
from abismal.distributions import FoldedNormal,Rice
from tensorflow_probability import distributions as tfd
from tensorflow_probability import util as tfu
from tensorflow_probability import bijectors as tfb
import tf_keras as tfk
from abismal.symmetry import Op


def centric_wilson(epsilon, sigma=1.):
    return tfd.HalfNormal(tf.math.sqrt(epsilon * sigma))

def acentric_wilson(epsilon, sigma=1.):
    return tfd.Weibull(2., tf.math.sqrt(epsilon * sigma))

class WilsonPrior(tfk.layers.Layer):
    """Wilson's priors on structure factor amplitudes."""
    def __init__(self, centric, epsilon, sigma=1.):
        """
        Parameters
        ----------
        centric : array
            Floating point or boolean array with value 1/True for centric reflections and 0/False. for acentric.
        epsilon : array
            Floating point array with multiplicity values for each structure factor.
        sigma : float or array
            The Σ value for the wilson distribution. The represents the average intensity stratified by a measure
            like resolution. 
        """
        super().__init__()
        self.epsilon = np.array(epsilon, dtype=np.float32)
        self.centric = np.array(centric, dtype=bool)
        self.sigma = np.array(sigma, dtype=np.float32)

        self.p_centric = centric_wilson(self.epsilon, self.sigma)
        self.p_acentric = acentric_wilson(self.epsilon, self.sigma)

    def log_prob(self, x):
        """
        Parameters
        ----------
        x : tf.Tensor
            Array of structure factor values with the same shape epsilon and centric.
        """
        return tf.where(self.centric, self.p_centric.log_prob(x), self.p_acentric.log_prob(x))

    def prob(self, x):
        """
        Parameters
        ----------
        x : tf.Tensor
            Array of structure factor values with the same shape epsilon and centric.
        """
        return tf.where(self.centric, self.p_centric.prob(x), self.p_acentric.prob(x))

    def mean(self):
        return tf.where(self.centric, self.p_centric.mean(), self.p_acentric.mean())

    def stddev(self):
        return tf.where(self.centric, self.p_centric.stddev(), self.p_acentric.stddev())

    def sample(self, *args, **kwargs):
        return tf.where(
            self.centric, 
            self.p_centric.sample(*args, **kwargs),
            self.p_acentric.sample(*args, **kwargs),
        )

class MultiWilsonPrior(tfk.layers.Layer):
    """
    This class uses reparameterized samples to approximate the log probability 
    of a multivariate Wilson prior. For this object, the user needs to specify
    a ReciprocalASUCollection instance which enumerates the sets of merged
    structure factors which will be produced. Additionally, a list of 
    parent asu identifiers will be provided. Each ASU can have at most one
    parent. The strength of the relationship between an ASU and its parent
    is given as a prior correlation coefficient. 

    Both the Single and Multi-Wilson prior can be expressed in terms of 
    folded normal and Ricean distributions. This is an important detail 
    that simplifies the implementation. 

    ```
    Wilson(F_h) = 
        FoldedNormal(0, sqrt(epsilon_h * Sigma_h)) #centric
        Rice(0., sqrt(0.5 * epsilon_h * Sigma_h)   #acentric
    ```
    where epsilon is the multiplicity and Sigma is the average reflection
    intensity. 

    ```
    DoubleWilson(F_h) = 
        FoldedNormal(r_h * z_Pa(h), sqrt(epsilon_h * Sigma_h * (1 - r^2))) #centric
        Rice(r_h * z_Pa(h), sqrt(0.5 * epsilon_h * Sigma_h * (1 - r^2)))   #acentric
    ```

    """
    def __init__(self, rac, parents, correlations, reindexing_ops=None, sigma=1., **kwargs):
        """
        Parameters
        ----------
        rac : ReciprocalASUCollection
            The reciprocal asu collection describing the merged structure factors
        parents : list
            An iterable of zero-indexed parents for each asu in the rac. Use None
            to indicate root nodes.
        correlations : list
            An iterable of prior correlation coefficients between asus. Use
            0.0 for root nodes. 
        reindexing_ops : list (optional)
            Optionally provide a list of reindexing operator strings, one per asu. 
        sigma : float or tensor (optional)
            Optionally provide an average intensity value for the prior. 
            If this is a tensor, it should have the combinded length of all
            the asus in the rac. 
        """
        super().__init__(**kwargs)
        #Store these for config purposes
        self._rac = rac.get_config()
        self._parents = parents
        self._correlations = correlations
        self._reindexing_ops = reindexing_ops
        self._sigma = sigma

        self.centric = rac.centric
        self.epsilon = rac.epsilon

        parent_ids = []
        is_root = []

        for asu_id, rasu in enumerate(rac):
            pa = parents[asu_id]
            if pa == asu_id or pa < 0:
                pa = None
            if pa is None:
                r = 0.
            else:
                r = correlations[asu_id]

            op = 'x,y,z'
            if reindexing_ops is not None:
                op = reindexing_ops[asu_id]
            if not isinstance(op, Op):
                op = Op(op)

            hkl = rasu.Hunique
            hkl = op(hkl)

            if pa is None:
                parent_id = -tf.ones(rasu.asu_size, rasu.miller_id.dtype)
                is_root.append(
                    tf.ones(rasu.asu_size, dtype='bool')
                )
            else:
                parent_id = rac._miller_ids(
                    pa * tf.ones_like(hkl[:,:1]),
                    hkl,
                )
                is_root.append(
                    tf.zeros(rasu.asu_size, dtype='bool')
                )
            parent_ids.append(parent_id)

        self.is_root = tf.concat(is_root, axis=0)
        self.parent_ids = tf.concat(parent_ids, axis=0)
        self.sigma = sigma

        idx = self.parent_ids >= 0
        r = tf.gather(tf.convert_to_tensor(correlations), rac.asu_id)
        self.r = tf.where(idx, r, 0.)

        self.scale = tf.where(
            rac.centric,
            tf.sqrt(rac.epsilon * sigma * (1. - tf.square(self.r))),
            tf.sqrt(0.5 * rac.epsilon * sigma * (1. - tf.square(self.r))),
        )
        self.has_parent = self.parent_ids >= 0
        self.parent_ids = tf.where(self.has_parent, self.parent_ids, tf.range(rac.asu_size, dtype=tf.int32))
        self.built = True #This is always true

        self.p_centric  = centric_wilson(self.epsilon, sigma)
        self.p_acentric = acentric_wilson(self.epsilon, sigma)

    @property
    def rac(self):
        return ReciprocalASUCollection.from_config(self._rac)

    def get_config(self):
        config = super().get_config()
        config.update({
            'rac' : self.rac,
            'parents' : self._parents,
            'correlations' : self._correlations,
            'reindexing_ops' : self._reindexing_ops,
            'sigma' : self.sigma,
        })
        return config

    def mean(self):
        """
        This is only for initialization of the surrogate!
        """
        return WilsonPrior(self.rac, self.sigma).mean()

    def log_prob(self, z):
        scale = self.scale #This is precomputed
        loc = self.r * tf.gather(z, self.parent_ids, axis=-1)

        # Mask any root nodes or missing parents
        loc = tf.where(self.has_parent, loc, 0.)

        ll  = tf.where(
            self.centric,
            FoldedNormal(loc, scale).log_prob(z),
            Rice(loc, scale).log_prob(z),
        )
        wilson_p = tf.where(
            self.centric,
            self.p_centric.log_prob(z),
            self.p_acentric.log_prob(z),
        )

        ll = tf.where(
            self.is_root,
            wilson_p,
            ll,
        )

        return ll






title = "phenix"
description = "Arguments for running PHENIX periodically during optimization."


args_and_kwargs=(
    (
        (
            "--eff-files",
        ),{
        "help": "Comma separated list of eff files.",
        "default": None,
        "type": str,
        }
    ),

    (
        (
            "--phenix-frequency",
        ),{
            "type": int,
            "default" : 1, 
            "help": 'How often to run phenix in epochs (default=1).',
        }
    ),
)



#/usr/bin/env cctbx.python
import numpy as np
import argparse
import reciprocalspaceship as rs
import pandas as pd
import gemmi
import tensorflow as tf
from abismal.io.loader import DataLoader
from reciprocalspaceship.decorators import spacegroupify,cellify
from multiprocessing import cpu_count,Pool




class StreamLoader(rs.io.crystfel.StreamLoader):
    @cellify
    @spacegroupify
    def __init__(self, stream_file, cell=None, dmin=None, asu_id=0, wavelength=None, encoding='utf-8'):
        """
        Parameters
        ----------
        stream_file : str
        cell : gemmi.UnitCell (optional)
            It is important that a consistent cell is used for resolution calculations. 
            If a cell is supplied it will be used. Otherwise, this class will first try to 
            located a unit cell record in the stream file header. Failing that, the average
            cell across the data set will be calculated. If a lattice type is availabe, the
            symmetry of the lattice will be used to correct cell axis lengths and angles. 
        dmin : float (optional)
        asu_id : int (optional)
            asu_id will be passed to the downstream model and is used to control which reflections 
            are merged together. 
        wavelength : float (optional)
            Optionally override the wavelengths recorded for each event in the stream file. 
        encoding : str (optional)
            Try to override the default encoding, 'utf-8'
        """
        super().__init__(stream_file, encoding)

        self.signature = None
        self.cell = self._get_unit_cell(cell)
        self.asu_id = asu_id
        self.wavelength = wavelength
        self.dmin = dmin

    def _get_unit_cell(self, cell=None):
        if cell is None:
            cell = self.extract_target_unit_cell()
            
        if cell is None:
            cell = self.calculate_average_unit_cell()
        if not isinstance(cell, gemmi.UnitCell):
            cell = gemmi.UnitCell(*cell)
        return cell

    def _convert_to_tf(self, peak_list, wavelength):
        hkl = peak_list[:,:3]
        d = self.cell.calculate_d_array(hkl)
        I = peak_list[:,3]
        SigI = peak_list[:,4]
        metadata = peak_list[:,5:]

        if self.dmin is not None:
            idx = d >= self.dmin
            d = d[idx]
            hkl = hkl[idx]
            peak_list = peak_list[idx]
            I = I[idx]
            SigI = SigI[idx]
            metadata = metadata[idx]

        I = tf.convert_to_tensor(I[None,:,None])
        d = tf.convert_to_tensor(d[None,:,None])
        SigI = tf.convert_to_tensor(SigI[None,:,None])
        hkl = tf.convert_to_tensor(hkl[None,:,:], dtype='int32')
        metadata = tf.convert_to_tensor(metadata[None,:,:])
        asu = tf.ones_like(I, dtype='int32') * self.asu_id
        wavelength = tf.ones_like(I) * wavelength

        data = ((asu, hkl, d, wavelength, metadata, I, SigI), (I,))
        return data

    def _parse_chunk(self, *args, **kwargs):
        data = super()._parse_chunk(*args, **kwargs)
        return [self._convert_to_tf(pl, data['wavelength']) for pl in data['peak_lists']]

    def get_dataset(self, peak_list_columns=None, **ray_kwargs):
        """
        Convert CrystFEL .stream files to a tf.data.Dataset.
        """

        if peak_list_columns is None:
            peak_list_columns = [
                'H', 'K', 'L', 
                'I', 'SigI', 
                's1x', 's1y', 's1z', 
                'ewald_offset_x', 'ewald_offset_y', 'ewald_offset_z',
            ]

        chunks = self.read_crystfel(
            wavelength=self.wavelength,
            peak_list_columns=peak_list_columns,
            **ray_kwargs,
        )

        def data_gen():
            for chunk in chunks:
                for peak_list in chunk:
                    yield peak_list

        if self.signature is None:
            for datum in data_gen():
                self.signature = StreamLoader.get_signature_from_datum(datum)
                break

        return tf.data.Dataset.from_generator(data_gen, output_signature=self.signature).unbatch()

    @staticmethod
    def get_signature_from_datum(datum):
        """Work out the proper signature for creating a dataset from an example"""
        #if ragged:
        #    signature = tf.nest.map_structure(tf.RaggedTensorSpec.from_value, datum)
        #else:
        #    signature = tf.nest.map_structure(tf.TensorSpec.from_tensor, datum)
        def to_ragged_spec(tensor):
            spec = tf.TensorSpec.from_tensor(tensor)
            shape,dtype = spec.shape,spec.dtype
            new = tf.TensorSpec(
                shape = (shape[0], None, shape[2]),
                dtype=dtype,
            )
            return new
        signature = tf.nest.map_structure(to_ragged_spec, datum)
        return signature

if __name__ == '__main__':
    file = "/mnt/raid/data/xtal/abismal_examples/cxidb_62/all-amb.stream.bz2"

    dmin = 1.5
    sg = "P 65"
    loader = StreamLoader(file, dmin=dmin, spacegroup=sg)
    ds = loader.get_dataset()
    count_max = 100
    count = 0
    for datum in ds:
        count += 1
        if count > count_max:
            break

    from IPython import embed
    embed(colors='linux')



import numpy as np
import tensorflow as tf
from tensorflow_probability import distributions as tfd


def WilsonPrior(centric, epsilon, sigma=1.):
    concentration = tf.where(centric, 0.5, 1.)
    rate = tf.where(centric, 0.5 / sigma / epsilon, 1. / sigma / epsilon)
    return tfd.Gamma(concentration, rate)




import reciprocalspaceship as rs
import tensorflow as tf
import tf_keras as tfk
from os.path import exists,dirname,abspath
from os import mkdir

class MtzSaver(tfk.callbacks.Callback):
    def __init__(self, output_directory, anomalous=True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.output_directory = abspath(output_directory)

        if not exists(self.output_directory):
            mkdir(output_directory)

    def on_epoch_end(self, epoch, logs=None):
        self.save_mtz(epoch)

    def save_mtz(self, epoch, seen=True):
        for asu_id,data in enumerate(self.model.surrogate_posterior.to_datasets(seen=seen)):
            data.write_mtz(f"{self.output_directory}/asu_{asu_id}_epoch_{epoch+1}.mtz")





import tensorflow as tf
from tensorflow_probability.python.internal import special_math
from tensorflow_probability import distributions as tfd
from tensorflow_probability.python.internal import dtype_util
import tensorflow_probability as tfp
from tensorflow_probability import bijectors as tfb
from tensorflow_probability.python.internal import tensor_util
import numpy as np
from tensorflow_probability.python.distributions import distribution
from tensorflow_probability.python.internal import reparameterization
from tensorflow_probability.python.distributions import kullback_leibler
from tensorflow_probability.python.internal import parameter_properties
from tensorflow_probability.python.bijectors import exp as exp_bijector

class Rice(distribution.AutoCompositeTensorDistribution):
    def __init__(self,
		   nu,
		   sigma,
		   validate_args=False,
		   allow_nan_stats=True,
		   name='Rice'):

        parameters = dict(locals())
        #Value of nu/sigma for which, above with the pdf and moments will be swapped with a normal distribution
        self._normal_crossover = 40. 
        dtype = dtype_util.common_dtype([nu, sigma], dtype_hint=tf.float32)
        with tf.name_scope(name) as name:
            self._nu = tensor_util.convert_nonref_to_tensor(nu)
            self._sigma = tensor_util.convert_nonref_to_tensor(sigma)
        super(Rice, self).__init__(
            dtype=dtype,
            validate_args=validate_args,
            allow_nan_stats=allow_nan_stats,
            reparameterization_type=reparameterization.FULLY_REPARAMETERIZED,
            parameters=parameters,
            name=name)

    @classmethod
    def _parameter_properties(cls, dtype, num_classes=None):
      # pylint: disable=g-long-lambda
        default_bij = lambda: exp_bijector.Exp()
        pp = dict(
            nu=parameter_properties.ParameterProperties(
                default_constraining_bijector_fn=default_bij
            ),
            sigma=parameter_properties.ParameterProperties(
                default_constraining_bijector_fn=default_bij
            )
        )
        return pp
      # pylint: enable=g-long-lambda

    @property
    def nu(self):
        return self._nu

    @property
    def sigma(self):
        return self._sigma

    @property
    def _pi(self):
        return tf.convert_to_tensor(np.pi, self.dtype)

    def sample_square(self, sample_shape=(), seed=None, name='sample', **kwargs):
        base_normal = tfd.Normal(0., self.sigma)
        s1 = base_normal.sample(sample_shape=sample_shape, seed=seed, name=name, **kwargs)
        s2 = base_normal.sample(sample_shape=sample_shape, seed=seed, name=name, **kwargs)
        return tf.square(s1) + tf.square(s2 + self.nu)

    def _sample_n(self, n, seed=None):
        sqr = self.sample_square(n, seed)
        return tf.math.sqrt(sqr)

    def _log_bessel_i0(self, x):
        return tf.math.log(tf.math.bessel_i0e(x)) + tf.math.abs(x)

    def _log_bessel_i1(self, x):
        return tf.math.log(tf.math.bessel_i1e(x)) + tf.math.abs(x)

    def _bessel_i0(self, x):
        return tf.math.exp(self._log_bessel_i0(x))

    def _bessel_i1(self, x):
        return tf.math.exp(self._log_bessel_i1(x))

    def _laguerre_half(self, x):
        return (1. - x) * tf.math.exp(x / 2. + self._log_bessel_i0(-0.5 * x)) - x * tf.math.exp(x / 2.  + self._log_bessel_i1(-0.5 * x) )

    def _prob(self, X):
        return tf.math.exp(self.log_prob(X))

    def _log_prob(self, X):
        sigma = self.sigma
        nu = self.nu
        log_p = tf.math.log(X) - 2.*tf.math.log(sigma) - (tf.square(X) + tf.square(nu))/(2*tf.square(sigma)) + \
                    self._log_bessel_i0(X * nu/tf.square(sigma))
        return tf.where(X <= 0, -np.inf, log_p)

    def _mean(self):
        sigma = self.sigma
        nu = self.nu
        snr = nu / sigma
        mean = sigma * tf.math.sqrt(self._pi / 2.) * self._laguerre_half(-0.5*tf.square(snr))
        return tf.where(snr > self._normal_crossover,  nu, mean)

    def _variance(self):
        sigma = self.sigma
        nu = self.nu
        snr = nu / sigma
        variance = 2*tf.square(sigma) + tf.square(nu) - 0.5*self._pi * tf.square(sigma) * tf.square(self._laguerre_half(-0.5*tf.square(snr)))
        return tf.where(snr > self._normal_crossover,  tf.square(sigma), variance)

    def _stddev(self):
        return tf.math.sqrt(self.variance())

    @property
    def _bivariate_normal(self):
        loc   = tf.convert_to_tensor([1., 0.], dtype=self.dtype)[...,None,:] * self.nu[...,None] 
        scale = tf.convert_to_tensor([1., 1.], dtype=self.dtype)[...,None,:] * self.sigma[...,None] 
        mvn = tfd.MultivariateNormalDiag(loc, scale)
        return mvn

@kullback_leibler.RegisterKL(Rice, Rice)
def _kl_rice_rice(q, p, name=None):
    return q._bivariate_normal.kl_divergence(p._bivariate_normal)

if __name__=="__main__":
    n = 100
    loc,scale = np.random.random((2, n)).astype('float32')
    q = Rice(loc, scale)
    p = Rice(0., 1.)

    x = np.linspace(-10., 10., 1000)
    u = q.mean()
    s = q.stddev()
    v = q.variance()

    q.log_prob(x[:,None])
    q.prob(x[:,None])
    q.sample(n)

    from IPython import embed
    embed(colors='linux')



from abismal.surrogate_posterior import IntensityPosteriorBase
from abismal.surrogate_posterior.intensity.wilson import WilsonPrior
import numpy as np
import reciprocalspaceship as rs
import tensorflow as tf
from tensorflow_probability import distributions as tfd
from tensorflow_probability import math as tfm
from tensorflow_probability import layers  as tfl
from tensorflow_probability import util as tfu
from tensorflow_probability import bijectors as tfb
import tf_keras as tfk


class GammaPosterior(IntensityPosteriorBase):
    def __init__(self, rac, kl_weight, scale_factor=1e-2, eps=1e-12, concentration_min=1., **kwargs):
        super().__init__(rac, **kwargs)
        self.rac = rac

        self.kl_weight = kl_weight
        loc = self.flat_prior().mean()
        scale = scale_factor * loc

        rate_init = tf.square(loc / scale)
        conc_init = loc / tf.square(scale)

        self.rate = tfu.TransformedVariable(
            rate_init,
            tfb.Chain([
                tfb.Shift(eps), 
                #tfb.Softplus(),
                tfb.Exp(),
            ]),
        )

        #Concentration should remain above one to prevent change in curvature
        self.concentration = tfu.TransformedVariable(
            conc_init,
            tfb.Chain([
                tfb.Shift(concentration_min + eps), 
                #tfb.Shift(eps), 
                #tfb.Softplus(),
                tfb.Exp(),
            ]),
        )

    def flat_prior(self):
        prior = WilsonPrior(
            self.rac.centric,
            self.rac.epsilon,
        )
        return prior

    def flat_distribution(self):
        return tfd.Gamma(self.concentration, self.rate)

# This is experimental still. 
#    def get_flat_fsigf(q, eps=1e-6):
#        """
#        Compute the mean and standard deviation of the square root of a gamma distribution.
#        """
#        alpha = q.concentration
#        beta  = q.rate
#        omega = alpha / beta
#
#        log_mean_sqrt_beta = tf.math.lgamma(alpha + 0.5) - tf.math.lgamma(alpha) #log(mean * beta**0.5)
#        mean = tf.math.exp(log_mean_sqrt_beta) / tf.sqrt(beta)
#
#        num = alpha - tf.math.exp(2.* log_mean_sqrt_beta)
#        var = tf.where(
#            alpha > 1e4,
#            0.25/beta, #The limit is 0.25 for num
#            num/beta,
#        )
#
#        std = tf.math.sqrt(var)
#        return mean,std



if __name__=="__main__":
    def sqrt_moments(gamma):
        k = alpha = gamma.concentration
        beta  = gamma.rate
        theta = tf.math.reciprocal(beta)

        """
        Random wikipedia knowledge:
        X ~ Gamma(k, θ), then \sqrt{X} follows a generalized gamma distribution with parameters 
        p = 2, d = 2k, and a = \sqrt{\theta}
        """
        p = 2.
        d = 2. * k
        a = tf.sqrt(theta)
        la = tf.math.log(a)
        log_mean = la + tf.math.lgamma((d + 1) / p) - tf.math.lgamma(d / p)
        log_t = 2.*la + tf.math.lgamma((d + 2) / p) - tf.math.lgamma(d / p)
        var = tf.exp(log_t) - tf.exp(2.*log_mean)
        std = tf.sqrt(var)
        mean = tf.exp(log_mean)

        return tf.exp(log_mean), tf.sqrt(var)


    n = 1000
    loc = np.random.random(n)
    scale = np.random.random(n)
    #scale = 0.0001 * loc

    rate = tf.square(loc / scale)
    conc = loc / tf.square(scale)

    q = tfd.Gamma(conc, rate)
    m,s = get_flat_fsigf(q)
    m,s = sqrt_moments(q)



title = "Training options"
description = "Options that deal with the specifics of model training"


args_and_kwargs = (
    (
        (
            "--mc-samples",
        ),{
            "help": "The number of monte carlo samples used to estimate gradients with default 256.",
            "default": 256,
            "type": int,
        }
    ),

    (
        (
            "--steps-per-epoch",
        ),{
            "help": "The number of gradient steps in an epoch dictating how often output can be saved. 1000 is the default.",
            "default": 1_000,
            "type": int,
        }
    ),

    (
        (
            "--epochs",
        ),{
            "help": "The number of training epochs to run with default 30.",
            "default": 30,
            "type": int,
        }
    ),

    (
        (
            "--validation-steps",
        ),{
            "help": "The number of validation steps run at the close of each epoch with default 100.",
            "default": 100,
            "type": int,
        }
    ),

    (
        (
            "--sample-reflections-per-image",
        ),{
            "help": "The number of reflections which will be sampled in order to calculate the image representation with default 32.",
            "default": 32,
            "type": int,
        }
    ),

    (
        (
            "--test-fraction",
        ),{
            "help": "The fraction of images reserved for validation with default 0.01.",
            "default": 0.01, "type": float,
        }
    ),

    (
        (
            "--shuffle-buffer-size",
        ),{
            "help": "The size of the shuffle buffer which randomizes the training data with default 100_000.",
            "default": 100_000,
            "type": int,
        }
    ),

    (
        (
            "--batch-size",
        ),{
            "help": "The size (number of images) in each training batch",
            "default": 100,
            "type": int,
        }
    ),

    (
        (
            "--disable-index-disambiguation",
        ),{
            "help": "Disable index disambiguation if applicable to the space group.",
            "action": "store_true",
        }
    ),


)



import tensorflow as tf
import tf_keras as tfk
from tensorflow_probability import distributions as tfd

def weighted_pearsonr(x, y, w=None, axis=-1, keepdims=False, eps=1e-12):
    """
    Calculate a [weighted Pearson correlation coefficient](https://en.wikipedia.org/wiki/Pearson_correlation_coefficient#Weighted_correlation_coefficient).

    Note
    ----
    x, y, and w may have arbitrarily shaped leading dimensions. The correlation coefficient will always be computed pairwise along the last axis.

    Parameters
    ----------
    x : np.array(float)
        An array of observations.
    y : np.array(float)
        An array of observations the same shape as x.
    w : np.array(float)
        An array of weights the same shape as x. These needn't be normalized.

    Returns
    -------
    r : float
        The Pearson correlation coefficient along the last dimension. This has shape {x,y,w}.shape[:-1].
    """
    if w is None:
        w = tf.ones_like(x)

    z = tf.math.reciprocal(tf.reduce_sum(w, axis=axis, keepdims=True))
    mx = tf.reduce_sum(z * (w * x), axis=axis, keepdims=True)
    my = tf.reduce_sum(z * (w * y), axis=axis, keepdims=True)

    dx = x - mx
    dy = y - my

    cxy = z * tf.reduce_sum(w * dx * dy, axis=axis, keepdims=True)
    cx = z * tf.reduce_sum(w * dx * dx, axis=axis, keepdims=True)
    cy = z * tf.reduce_sum(w * dy * dy, axis=axis, keepdims=True)

    r = cxy / tf.sqrt(cx * cy + eps)
    return r

class LocationScale(tfk.layers.Layer):
    def _likelihood(self, iobs, sigiobs):
        raise NotImplementedError(
            "Derived classes must implement _likelihood(iobs, sigiobs)->tfd.Distribution"
        )

    def register_metrics(self, ipred, iobs, sigiobs):
        likelihood = self._likelihood(iobs, sigiobs)

        # This is the mean ipred across the posterior mc samples
        iobs = iobs * tf.ones_like(ipred)
        sigiobs = sigiobs * tf.ones_like(ipred)
        w = tf.math.reciprocal(tf.square(sigiobs))

        cc = weighted_pearsonr(iobs, ipred, w, axis=(-2, -1))
        cc = tf.squeeze(cc)
        self.add_metric(cc, name='wCCpred')

        cc = weighted_pearsonr(iobs, ipred, axis=(-2, -1))
        cc = tf.squeeze(cc)
        self.add_metric(cc, name='CCpred')

        cdf = likelihood.cdf(ipred)
        f = tf.reduce_mean(
            tf.where(
                (cdf >= 0.025)&(cdf <= 0.975),
                1.,
                0.,
            )
        )
        self.add_metric(f, "C95")

        resid = ipred - iobs
        r2 = tf.square(resid)

        #mse = tf.reduce_mean(r2)
        #self.add_metric(mse, name='MSE')

        #mae = tf.reduce_mean(tf.abs(resid))
        #self.add_metric(mse, name='MAE')

        #wmse = tf.reduce_sum(w * r2) / tf.reduce_sum(w * tf.ones_like(r2))
        #self.add_metric(wmse, name='WMSE')

        #z = tf.abs(resid / sigiobs) 
        #z1 = tf.reduce_mean(tf.where(z > 1., 0., 1.))
        #z2 = tf.reduce_mean(tf.where(z > 2., 0., 1.))
        #z3 = tf.reduce_mean(tf.where(z > 3., 0., 1.))
        #self.add_metric(z1, name='Z1_Frac')
        #self.add_metric(z2, name='Z2_Frac')
        #self.add_metric(z3, name='Z3_Frac')

    def call(self, ipred, iobs, sigiobs):
        likelihood = self._likelihood(iobs, sigiobs)
        ll = likelihood.log_prob(ipred)
        return ll




#/usr/bin/env cctbx.python
import numpy as np
import argparse
import reciprocalspaceship as rs
import pandas as pd
import gemmi
import tensorflow as tf
from .loader import DataLoader


def get_first_key_of_type(ds, dtype):
    idx = ds.dtypes == dtype
    if idx.sum() == 0:
        raise ValueError(f"Dataset has no key of type {dtype}")
    key = ds.dtypes[idx].keys()[0]
    return key

class MTZLoader(DataLoader):
    def __init__(
            self, 
            mtz_file, 
            wavelength=1.,
            dmin=None,
            metadata_keys=None, 
            cell=None, 
            spacegroup=None, 
            batch_key=None, 
            intensity_key=None,
            sigma_key=None,
            asu_id=0,
        ):
        self.metadata_keys = metadata_keys
        if self.metadata_keys is None:
            self.metadata_keys = [
                "XDET",
                "YDET",
            ]

        super().__init__(len(self.metadata_keys))

        self.wavelength = wavelength
        self.mtz_file = mtz_file
        self.dmin = dmin
        self.cell = cell
        self.spacegroup = spacegroup
        self.asu_id = asu_id

        self.batch_key = batch_key
        self.intensity_key = intensity_key
        self.sigma_key = sigma_key

    def get_dataset(self):
        ds = rs.read_mtz(self.mtz_file)

        if self.cell is None:
            self.cell = ds.cell
        else:
            ds.cell = self.cell

        if self.spacegroup is None:
            self.spacegroup = ds.spacegroup
        else:
            ds.spacegroup = self.spacegroup

        if self.batch_key is None:
            self.batch_key = get_first_key_of_type(ds, 'B')
        if self.intensity_key is None:
            self.intensity_key = get_first_key_of_type(ds, 'J')
        if self.sigma_key is None:
            self.sigma_key = get_first_key_of_type(ds, 'Q')


        ds.compute_dHKL(True).label_absences(True)
        ds = ds[~ds.ABSENT]
        if self.dmin is not None:
            ds = ds[ds.dHKL >= self.dmin]
        ds['rowids'] = ds.groupby(self.batch_key).ngroup().to_numpy("int32")
        ds.sort_values("rowids", inplace=True)
        rowids = ds.rowids.to_numpy("int32")

        I = tf.RaggedTensor.from_value_rowids(ds[self.intensity_key].to_numpy("float32")[:,None], rowids)
        d = tf.RaggedTensor.from_value_rowids(ds['dHKL'].to_numpy("float32")[:,None], rowids)
        SigI = tf.RaggedTensor.from_value_rowids(ds[self.sigma_key].to_numpy("float32")[:,None], rowids)
        Metadata = tf.RaggedTensor.from_value_rowids(ds[self.metadata_keys].to_numpy("float32"), rowids)
        HKL = tf.RaggedTensor.from_value_rowids(ds.get_hkls(), rowids)
        wavelength = tf.ones_like(I) * self.wavelength
        ASU = tf.ones_like(I, dtype='int32') * self.asu_id

        tfds = tf.data.Dataset.from_tensor_slices(
            ((ASU, HKL, d, wavelength, Metadata, I, SigI), (I,)),
        )

        return tfds

if __name__=='__main__':
    stream_file = '/mnt/raid/data/xtal/20210615_Neutze_collab/indexing_dark_before_merging_intensities.stream'
    test = np.random.random((1_000, 6))
    mean, variance = StreamLoader.online_mean_variance(test)
    assert np.all(np.isclose(mean, test.mean(0)))
    assert np.all(np.isclose(variance, np.square(test.std(0))))

    for crystal in StreamLoader.to_blocks(stream_file, 'crystal'):
        break

    assert len(crystal) > 0

    cell = StreamLoader.get_average_cell(stream_file)

    geo = StreamLoader.to_blocks(stream_file, 'geometry')
    print(''.join(next(geo)))

    print(StreamLoader.get_wavelength(stream_file))



title = "IO"
description = "Arguments controlling file inputs and outputs."


args_and_kwargs=(
    (
        (
            "-o",
            "--out-dir",
        ),{
        "help": "The directory in which to output results. The current working directory by default",
        "default": ".",
        "type": str,
        }
    ),

    (
        (
            "--num-cpus",
        ),{
        "help": "Number of CPUs to use for parsing CrystFEL .stream files with default 1.",
        "default": 1,
        "type": int,
        }
    ),

    (
        (
            "inputs",
        ),{
            "nargs": '+',
            "help": 'Either .stream files from CrystFEL or .refl and .expt files from dials',
        }
    ),

    (
        (
            "--wavelength",
        ),{
            "type": float,
            "default" : None,
            "help": 'Override the wavelengths inferred from the inputs.',
        }
    ),

    (
        (
            "--separate",
        ),{
        "help": "Merge the contents of each input file into a separate output file. ",
        "action": "store_true",
        }
    ),
)



title = "Likelihood"
description = "Arguments affecting the likelihood term in the ELBO estimated by ABISMAL"

args_and_kwargs=(
    (
        (
            "-t",
            "--studentt-dof",
        ),{
            "help": "Use a t-distributed error model with this many degrees of freedom.",
            "type" : float,
            "default": None,
        }
    ),
)



title = "Priors"
description = "Arguments governing the prior distributions"

from abismal.command_line.parser.custom_types import list_of_ints,list_of_ops,list_of_floats

args_and_kwargs = (
    (
        (
            "--kl-weight",
        ),{
            "help": "The strength of the structure factor prior distribution with default 1.0.",
            "default": 1.0,
            "type": float,
        }
    ),

    (
        (
            "--scale-kl-weight",
        ),{
            "help": "The strength of the scale prior distribution with default 0.0.",
            "default": 0.0,
            "type": float,
        }
    ),

    (
        (
            "--parents",
        ),{
            "help": "Set parent asu for the Multi-Wilson prior. This is used with --separate flag." 
                    "Supply parent asu ids as a comma separated list of integers. If an asu has no"
                    " parent, supply its own asu id. for example, --parents 0,0, indicates that "
                    "the first asu has no parent and the second asu is dependent on the first.",
            "default": None,
            "type": list_of_ints,
        }
    ),

    (
        (
            "-r",
            "--prior-correlation",
        ),{
            "help": "The prior correlation (r-value) for each ASU and its parent. Supply "
                    "comma-separated floating point Values. " 
                    "Values supplied for ASUs without a parent will be ignored. "
                    "Example: -r 0.0,0.99",
            "default": None,
            "type": list_of_floats,
        }
    ),

    (
        (
            "--reindexing-ops",
        ),{
            "help": 'Supply semicolon-separated reindexing ops which map from the child asu' 
                    'convention into the parent convention. ' 
                    'For example, --reindexing-ops "x,y,z;-x,-y,-z". ',
            "default": None,
            "type": list_of_ops,
        }
    ),
)



from abismal.distributions import FoldedNormal, Rice
import tensorflow as tf
from tensorflow_probability.python.internal import special_math
from tensorflow_probability import distributions as tfd
from tensorflow_probability.python.internal import dtype_util
import tensorflow_probability as tfp
from tensorflow_probability import bijectors as tfb
from tensorflow_probability.python.internal import tensor_util
import numpy as np
from tensorflow_probability.python.distributions import distribution
from tensorflow_probability.python.internal import reparameterization
from tensorflow_probability.python.distributions import kullback_leibler
from tensorflow_probability.python.internal import parameter_properties
from tensorflow_probability.python.bijectors import exp as exp_bijector

class RiceWoolfson(distribution.AutoCompositeTensorDistribution):
    def __init__(self,
		   loc,
		   scale,
           centric,
		   validate_args=False,
		   allow_nan_stats=True,
		   name='Rice'):

        parameters = dict(locals())
        dtype = dtype_util.common_dtype([loc, scale], dtype_hint=tf.float32)
        with tf.name_scope(name) as name:
            self._loc = tensor_util.convert_nonref_to_tensor(loc)
            self._scale = tensor_util.convert_nonref_to_tensor(scale)
            self._centric = tensor_util.convert_nonref_to_tensor(centric, dtype='bool')
        super(RiceWoolfson, self).__init__(
            dtype=dtype,
            validate_args=validate_args,
            allow_nan_stats=allow_nan_stats,
            reparameterization_type=reparameterization.FULLY_REPARAMETERIZED,
            parameters=parameters,
            name=name)

    @classmethod
    def _parameter_properties(cls, dtype, num_classes=None):
        # pylint: disable=g-long-lambda
        pp = dict(
            loc=parameter_properties.ParameterProperties(),
            scale=parameter_properties.ParameterProperties(
                default_constraining_bijector_fn=(
                    lambda: exp_bijector.Exp()
                )
            ),
            centric=parameter_properties.ParameterProperties(),
        )
        return pp
        # pylint: enable=g-long-lambda

    @property
    def loc(self):
        return self._loc

    @property
    def scale(self):
        return self._scale

    @property
    def centric(self):
        return self._centric

    @property
    def _centric_distribution(self):
        return FoldedNormal(self.loc, self.scale)

    @property
    def _acentric_distribution(self):
        return Rice(self.loc, self.scale)

    def _log_prob(self, X, **kwargs):
        return tf.where(
            self._centric,
            self._centric_distribution.log_prob(X),
            self._acentric_distribution.log_prob(X),
        )

    def _sample_n(self, n, seed=None):
        return tf.where(
            self._centric,
            self._centric_distribution.sample(n, seed=seed),
            self._acentric_distribution.sample(n, seed=seed),
        )

    def _mean(self):
        return tf.where(
            self._centric,
            self._centric_distribution.mean(),
            self._acentric_distribution.mean(),
        )

    def _variance(self):
        return tf.where(
            self._centric,
            self._centric_distribution.variance(),
            self._acentric_distribution.variance(),
        )

    def _stddev(self):
        return tf.where(
            self._centric,
            self._centric_distribution.stddev(),
            self._acentric_distribution.stddev(),
        )

    def sample_intensities(self, sample_shape=(), seed=None, name='sample', **kwargs):
        return tf.where(
            self.centric,
            self._centric_distribution.sample_square(sample_shape=sample_shape, seed=seed, name=name, **kwargs),
            self._acentric_distribution.sample_square(sample_shape=sample_shape, seed=seed, name=name, **kwargs),
        )

@kullback_leibler.RegisterKL(RiceWoolfson, RiceWoolfson)
def _kl_rice_rice_woolfson(q, p, name=None):
    return tf.where(
        q.centric,
        q._centric_distribution.kl_divergence(p._centric_distribution),
        q._acentric_distribution.kl_divergence(p._acentric_distribution),
    )



if __name__=="__main__":
    n = 100
    loc,scale = np.random.random((2, n)).astype('float32')
    centric = np.random.choice([False, True], size=n)
    q = RiceWoolfson(loc, scale, centric)
    p = RiceWoolfson(0., 1., centric)
    x = np.linspace(-10., 10., 1000)
    u = q.mean()
    s = q.stddev()
    v = q.variance()

    q.log_prob(x[:,None])
    q.prob(x[:,None])
    q.sample(n)
    q.sample_intensities(n)
    q.kl_divergence(p)

    from IPython import embed
    embed(colors='linux')




title = "Optimizer"
description = "Arguments affecting the optimization algorithm"


args_and_kwargs=(

    (
        (
            "--learning-rate",
        ),{
            "help": "Learning rate for Adam with default 1e-3.",
            "default": 1e-3,
            "type": float
        }
    ),

    (
        (
            "--learning-rate-final",
        ),{
            "help": "Optionally anneal the learning rate to this value throughout training.",
            "default": None,
            "type": float
        }
    ),

    (
        (
            "--beta-1",
        ),{
            "help": "First moment momentum parameter for Adam with default 0.9.",
            "default": 0.9,
            "type": float,
        }
    ),

    (
        (
            "--beta-2",
        ),{
            "help": "Second moment momentum parameter for Adam with default 0.999.",
            "default": 0.999,
            "type": float,
        }
    ),

    (
        (
            "--adam-epsilon",
        ),{
            "help": "A small constant for numerical stability with default 1e-9.",
            "default": 1e-9,
            "type": float
        }
    ),

    (
        (
            "--global-clipnorm",
        ),{
            "help": "Optionally apply gradient clipping with a global norm.",
            "default": None,
            "type": float,
        }
    ),

    (
        (
            "--clipnorm",
        ),{
            "help": "Optionally apply gradient clipping with a per-parameter norm.",
            "default": None,
            "type": float,
        }
    ),

    (
        (
            "--clip",
        ),{
            "help": "Optionally apply gradient clipping with a value.",
            "default": None,
            "type": float,
        }
    ),

    (
        (
            "--amsgrad",
        ),{
            "help": "Optionally use the amsgrad variant of Adam.",
            "action": 'store_true',
        }
    ),
)



# Copyright 2018 The TensorFlow Probability Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The Truncated Normal distribution class."""

# Dependency imports
import numpy as np

import tensorflow.compat.v2 as tf

from tensorflow_probability.python.bijectors import sigmoid as sigmoid_bijector
from tensorflow_probability.python.bijectors import softplus as softplus_bijector
from tensorflow_probability.python.distributions import distribution
from tensorflow_probability.python.internal import assert_util
from tensorflow_probability.python.internal import dtype_util
from tensorflow_probability.python.internal import parameter_properties
from tensorflow_probability.python.internal import prefer_static as ps
from tensorflow_probability.python.internal import reparameterization
from tensorflow_probability.python.internal import samplers
from tensorflow_probability.python.internal import special_math
from tensorflow_probability.python.internal import tensor_util
from tensorflow_probability.python.math.generic import log_sub_exp as _log_sub_exp


__all__ = [
    'TruncatedNormal',
]


def _normal_pdf(x):
  two_pi = tf.convert_to_tensor(2 * np.pi, dtype=x.dtype)
  return tf.math.rsqrt(two_pi) * tf.exp(-0.5 * tf.square(x))


def _normal_log_pdf(x):
  two_pi = tf.convert_to_tensor(2 * np.pi, dtype=x.dtype)
  return -0.5 * (tf.math.log(two_pi) + tf.square(x))


def _normal_cdf_difference(x, y):
  """Computes ndtr(x) - ndtr(y) assuming that x >= y."""
  # When x >= y >= 0, we will return ndtr(-y) - ndtr(-x)
  # because ndtr does not have a good precision for large positive x, y.
  is_y_positive = y >= 0
  x_hat = tf.where(is_y_positive, -y, x)
  y_hat = tf.where(is_y_positive, -x, y)
  return special_math.ndtr(x_hat) - special_math.ndtr(y_hat)


def _normal_cdf_log_difference(x, y):
  """Computes log(ndtr(x) - ndtr(y)) assuming that x >= y."""
  # When x >= y >= 0, we will return log(ndtr(-y) - ndtr(-x))
  # because ndtr does not have a good precision for large positive x, y.
  is_y_positive = y >= 0
  x_hat = tf.where(is_y_positive, -y, x)
  y_hat = tf.where(is_y_positive, -x, y)
  return _log_sub_exp(
      special_math.log_ndtr(x_hat), special_math.log_ndtr(y_hat))


class TruncatedNormal(distribution.AutoCompositeTensorDistribution):
  """The Truncated Normal distribution.

  The truncated normal is a normal distribution bounded between `low`
  and `high` (the pdf is 0 outside these bounds and renormalized).

  Samples from this distribution are differentiable with respect to `loc`,
  `scale` as well as the bounds, `low` and `high`, i.e., this
  implementation is fully reparameterized.

  For more details, see [here](
  https://en.wikipedia.org/wiki/Truncated_normal_distribution).

  ### Mathematical Details

  The probability density function (pdf) of this distribution is:
  ```none
    pdf(x; loc, scale, low, high) =
        { (2 pi)**(-0.5) exp(-0.5 y**2) / (scale * z) for low <= x <= high
        { 0                                    otherwise
    y = (x - loc)/scale
    z = NormalCDF((high - loc) / scale) - NormalCDF((low - loc) / scale)
  ```

  where:

  * `NormalCDF` is the cumulative density function of the Normal distribution
    with 0 mean and unit variance.

  This is a scalar distribution so the event shape is always scalar and the
  dimensions of the parameters define the batch_shape.

  #### Examples
  ```python

  tfd = tfp.distributions
  # Define a batch of two scalar TruncatedNormals with modes at 0. and 1.0
  dist = tfd.TruncatedNormal(loc=[0., 1.], scale=1.,
                             low=[-1., 0.],
                             high=[1., 1.])

  # Evaluate the pdf of the distributions at 0.5 and 0.8 respectively returning
  # a 2-vector tensor.
  dist.prob([0.5, 0.8])

  # Get 3 samples, returning a 3 x 2 tensor.
  dist.sample([3])
  ```
  """

  def __init__(self,
               loc,
               scale,
               low=0.,
               validate_args=False,
               allow_nan_stats=True,
               name='TruncatedNormal'):
    """Construct TruncatedNormal.

    All parameters of the distribution will be broadcast to the same shape,
    so the resulting distribution will have a batch_shape of the broadcast
    shape of all parameters.

    Args:
      loc: Floating point tensor; the mean of the normal distribution(s) (
        note that the mean of the resulting distribution will be different
        since it is modified by the bounds).
      scale: Floating point tensor; the std deviation of the normal
        distribution(s).
      low: `float` `Tensor` representing lower bound of the distribution's
        support. Must be such that `low < high`.
      validate_args: Python `bool`, default `False`. When `True` distribution
        parameters are checked at run-time.
      allow_nan_stats: Python `bool`, default `True`. When `True`,
        statistics (e.g., mean, mode, variance) use the value '`NaN`' to
        indicate the result is undefined. When `False`, an exception is raised
        if one or more of the statistic's batch members are undefined.
      name: Python `str` name prefixed to Ops created by this class.
    """
    parameters = dict(locals())
    high = np.inf
    with tf.name_scope(name) as name:
      dtype = dtype_util.common_dtype([loc, scale, low, high], tf.float32)
      self._loc = tensor_util.convert_nonref_to_tensor(
          loc, name='loc', dtype=dtype)
      self._scale = tensor_util.convert_nonref_to_tensor(
          scale, name='scale', dtype=dtype)
      self._low = tensor_util.convert_nonref_to_tensor(
          low, name='low', dtype=dtype)
      self._high = tensor_util.convert_nonref_to_tensor(
          high, name='high', dtype=dtype)
      dtype_util.assert_same_float_dtype(
          [self._loc, self._scale, self._low, self._high])

      super(TruncatedNormal, self).__init__(
          dtype=dtype,
          # This distribution is fully reparameterized. loc, scale have straight
          # through gradients. The gradients for the bounds are implemented
          # using custom derived expressions based on implicit gradients.
          # For the special case of lower bound zero and a positive upper bound
          # an equivalent expression can also be found in Sec 9.1.1.
          # of https://arxiv.org/pdf/1806.01851.pdf. The implementation here
          # handles arbitrary bounds.
          reparameterization_type=reparameterization.FULLY_REPARAMETERIZED,
          validate_args=validate_args,
          allow_nan_stats=allow_nan_stats,
          parameters=parameters,
          name=name)

  def _loc_scale_low_high(self, loc=None, scale=None, low=None, high=None):
    loc = tf.convert_to_tensor(self.loc if loc is None else loc)
    scale = tf.convert_to_tensor(self.scale if scale is None else scale)
    low = tf.convert_to_tensor(self.low if low is None else low)
    high = tf.convert_to_tensor(self.high if high is None else high)
    return loc, scale, low, high

  def _standardized_low_and_high(self,
                                 loc=None,
                                 scale=None,
                                 low=None,
                                 high=None):
    loc, scale, low, high = self._loc_scale_low_high(
        loc=loc, scale=scale, low=low, high=high)
    high = np.inf
    return (low - loc) / scale, (high - loc) / scale

  def _log_normalizer(self,
                      loc=None,
                      scale=None,
                      low=None,
                      high=None,
                      std_low=None,
                      std_high=None):
    if std_low is None or std_high is None:
      std_low, _ = self._standardized_low_and_high(
          loc=loc, scale=scale, low=low, high=high)
      std_high = np.inf
    result = _log_sub_exp(
      0., special_math.log_ndtr(std_low))
    return result
    #return _normal_cdf_log_difference(std_high, std_low)

  @classmethod
  def _parameter_properties(cls, dtype, num_classes=None):
    # pylint: disable=g-long-lambda
    return dict(
        loc=parameter_properties.ParameterProperties(),
        scale=parameter_properties.ParameterProperties(
            default_constraining_bijector_fn=(
                lambda: softplus_bijector.Softplus(low=dtype_util.eps(dtype)))),
        low=parameter_properties.ParameterProperties(),
        # TODO(b/169874884): Support decoupled parameterization.
        high=parameter_properties.ParameterProperties(
            default_constraining_bijector_fn=parameter_properties
            .BIJECTOR_NOT_IMPLEMENTED,))
    # pylint: enable=g-long-lambda

  @property
  def loc(self):
    return self._loc

  @property
  def scale(self):
    """Distribution parameter for the scale."""
    return self._scale

  @property
  def low(self):
    return self._low

  @property
  def high(self):
    return self._high

  def _event_shape(self):
    return tf.TensorShape([])

  def _sample_n(self, n, seed=None):
    loc, scale, low, high = self._loc_scale_low_high()
    batch_shape = self._batch_shape_tensor(
        loc=loc, scale=scale, low=low, high=high)
    sample_and_batch_shape = ps.concat([[n], batch_shape], 0)
    result = tf.random.stateless_parameterized_truncated_normal(
        shape=sample_and_batch_shape,
        means=loc,
        stddevs=scale,
        minvals=low,
        maxvals=high,
        seed=samplers.sanitize_seed(seed))
    result = tf.clip_by_value(result, low, high)
    return result

  def _log_prob(self, x):
    np_dtype = dtype_util.as_numpy_dtype(x.dtype)
    loc, scale, low, high = self._loc_scale_low_high()
    log_prob = -(np_dtype(0.5) * tf.square(
        (x - loc) / scale) + (0.5 * np.log(2. * np.pi)).astype(np_dtype) +
                 tf.math.log(scale) +
                 self._log_normalizer(loc=loc, scale=scale, low=low, high=high))
    # p(x) is 0 outside the bounds.
    bounded_log_prob = tf.where((x > high) | (x < low),
                                np_dtype(-np.inf),
                                log_prob)
    return bounded_log_prob

  def _cdf(self, x):
    return tf.exp(self._log_cdf(x))

  def _log_cdf(self, x):
    np_dtype = dtype_util.as_numpy_dtype(x.dtype)
    loc, scale, low, high = self._loc_scale_low_high()
    std_low, std_high = self._standardized_low_and_high(
        low=low, high=high, loc=loc, scale=scale)
    std_x = (x - loc) / scale
    log_cdf = (
        _normal_cdf_log_difference(std_x, std_low) -
        self._log_normalizer(std_low=std_low, std_high=std_high))
    # cdf(x) is 0 when x < low and is 1 when x > high.
    bounded_log_cdf = tf.where(x > high, np_dtype(0.), log_cdf)
    bounded_log_cdf = tf.where(x < low, np_dtype(-np.inf), bounded_log_cdf)
    return bounded_log_cdf

  def _entropy(self):
    loc, scale, low, high = self._loc_scale_low_high()
    std_low, std_high = self._standardized_low_and_high(
        loc=loc, scale=scale, low=low, high=high)
    log_normalizer = self._log_normalizer(std_low=std_low, std_high=std_high)
    return (
        0.5 * (1 + np.log(2.) + np.log(np.pi)) + tf.math.log(scale) +
        log_normalizer + 0.5 *
        (std_low * _normal_pdf(std_low) - std_high * _normal_pdf(std_high)) /
        tf.exp(log_normalizer))

  def _mean(self):
    loc, scale, low, high = self._loc_scale_low_high()
    std_low, std_high = self._standardized_low_and_high(
        loc=loc, scale=scale, low=low, high=high)
    lse, sign = _log_sub_exp(_normal_log_pdf(std_low),
                             _normal_log_pdf(std_high),
                             return_sign=True)
    return loc + scale * sign * tf.math.exp(
        lse - self._log_normalizer(std_low=std_low, std_high=std_high))

  def _mode(self):
    # mode = { loc: for low <= loc <= high
    #          low: for loc < low
    #          high: for loc > high
    #        }
    loc = tf.convert_to_tensor(self.loc)
    low = tf.convert_to_tensor(self.low)
    high = tf.convert_to_tensor(self.high)
    shp = self._batch_shape_tensor(loc=loc, low=low, high=high)
    # We *must* broadcast with scale to get a correctly shaped output, but
    # TODO(b/141460015): we should not have to explicitly broadcast the first
    # parameter to clip_by_value to align with the second and third parameters.
    bc_loc = tf.broadcast_to(loc, shp)
    return tf.clip_by_value(bc_loc, low, high)

  def _quantile(self, p):
    # TODO(b/188413116): This implementation is analytically correct, but might
    # not perform well in all cases. See
    # https://en.wikipedia.org/wiki/Truncated_normal_distribution#Generating_values_from_the_truncated_normal_distribution)
    # for a discussion on alternatives.
    loc, scale, low, high = self._loc_scale_low_high()
    std_low, std_high = self._standardized_low_and_high(
        low=low, high=high, loc=loc, scale=scale)
    quantile = tf.math.ndtri(
        special_math.ndtr(std_low) +
        p * _normal_cdf_difference(std_high, std_low)) * scale + loc
    return quantile

  def _variance(self):
    #Modified by kmd to respect high = np.inf
    loc, scale, low, high = self._loc_scale_low_high()
    std_low, std_high = self._standardized_low_and_high(
        loc=loc, scale=scale, low=low, high=high)
    log_normalizer = self._log_normalizer(std_low=std_low, std_high=std_high)

    var = (
        tf.square(scale) *
        (1. +
         (std_low * _normal_pdf(std_low)) /
         tf.exp(log_normalizer) -
         tf.exp(2. * (
             _log_sub_exp(  # ignore sign because result gets squared
                 _normal_log_pdf(std_low), _normal_log_pdf(std_high))
             - log_normalizer))))
    return var

  def _default_event_space_bijector(self):
    return sigmoid_bijector.Sigmoid(
        low=self.low, high=self.high, validate_args=self.validate_args)

  def _parameter_control_dependencies(self, is_init):
    if not self.validate_args:
      return []
    assertions = []
    low = None
    high = None
    if is_init != tensor_util.is_ref(self.low):
      low = tf.convert_to_tensor(self.low)
      assertions.append(
          assert_util.assert_finite(low, message='`low` is not finite'))
    if is_init != tensor_util.is_ref(self.high):
      high = tf.convert_to_tensor(self.high)
      assertions.append(
          assert_util.assert_finite(high, message='`high` is not finite'))
    if is_init != tensor_util.is_ref(self.loc):
      assertions.append(
          assert_util.assert_finite(self.loc, message='`loc` is not finite'))
    if is_init != tensor_util.is_ref(self.scale):
      scale = tf.convert_to_tensor(self.scale)
      assertions.extend([
          assert_util.assert_positive(
              scale, message='`scale` must be positive'),
          assert_util.assert_finite(scale, message='`scale` is not finite'),
      ])
    if (is_init != tensor_util.is_ref(self.low) or
        is_init != tensor_util.is_ref(self.high)):
      low = tf.convert_to_tensor(self.low) if low is None else low
      high = tf.convert_to_tensor(self.high) if high is None else high
      assertions.append(
          assert_util.assert_greater(
              high,
              low,
              message='TruncatedNormal not defined when `low >= high`.'))
    return assertions

  def _sample_control_dependencies(self, x):
    assertions = []
    if not self.validate_args:
      return assertions
    assertions.append(assert_util.assert_greater_equal(
        x, self.low, message='Sample must be greater than or equal to `low`.'))
    assertions.append(assert_util.assert_less_equal(
        x, self.high, message='Sample must be less than or equal to `high`.'))
    return assertions


if __name__=="__main__":
    n = 1_000
    loc, scale = np.random.random((2, n))
    from tensorflow_probability import distributions as tfd
    low,high = 0.,1e32
    q_true = tfd.TruncatedNormal(loc, scale, low, high)
    q_test = TruncatedNormal(loc, scale, low)

    assert np.allclose(q_true.mean(), q_test.mean())
    assert np.allclose(q_true.stddev(), q_test.stddev())

    s = 50
    z = q_test.sample(s)
    p_test = q_test.log_prob(z)
    p_true = q_true.log_prob(z)

    assert np.allclose(p_test, p_true)
    from IPython import embed
    embed(colors='linux')







#!/usr/bin/env python

def main():
    from abismal.command_line.parser import parser
    parser = parser.parse_args()

    from abismal.io.tf_settings import set_log_level, set_gpu
    set_log_level(parser.tf_log_level)
    set_gpu(parser.gpu_id)
    run_abismal(parser)


# TODO: refactor this filetype control flow into abismal.io
_file_endings = {
    'refl' : ('.refl', '.pickle'),
    'expt' : ('.expt', '.json'),
    'stream' : ('.stream',),
}

def _is_file_type(s, endings):
    for ending in endings:
        if s.endswith(ending):
            return True
    return False

def _is_stream_file(s):
    return _is_file_type(s, _file_endings['stream'])

def _is_refl_file(s):
    return _is_file_type(s, _file_endings['refl'])

def _is_expt_file(s):
    return _is_file_type(s, _file_endings['expt'])

def _is_dials_file(s):
    return _is_refl_file(s) or _is_expt_file(s)

def run_abismal(parser):
    from abismal import __version__ as version
    from abismal.symmetry import ReciprocalASU,ReciprocalASUCollection
    from abismal.merging import VariationalMergingModel
    from abismal.callbacks import HistorySaver,MtzSaver,PhenixRunner,AnomalousPeakFinder
    from abismal.io import split_dataset_train_test,set_gpu
    from abismal.scaling import ImageScaler
    from abismal.surrogate_posterior.structure_factor import FoldedNormalPosterior
    from abismal.surrogate_posterior.intensity import GammaPosterior
    from tf_keras.optimizers import Adam
    from tf_keras.callbacks import ModelCheckpoint
    import gemmi
    from tensorflow.data import AUTOTUNE
    import logging
    from os.path import exists
    from os import mkdir

    if not exists(parser.out_dir):
        mkdir(parser.out_dir)

    log_file = parser.out_dir + "/abismal.log"
    logger = logging.getLogger(__name__)
    logging.basicConfig(filename=log_file, level=logging.DEBUG)
    logger.info(f"Starting abismal, version {version}")
    logger.info("Running with the following options... ")
    for k,v in vars(parser).items():
        logger.info(f"{k} : {v}")
    logger.info(str(vars(parser)))

    cell = parser.cell
    space_group = parser.space_group
    asu_id = 0
    if all([_is_stream_file(f) for f in parser.inputs]):
        from abismal.io import StreamLoader
        data = None
        cell = parser.cell

        for stream_file in parser.inputs:
            loader = StreamLoader(
                stream_file, 
                cell=cell, 
                dmin=parser.dmin, 
                asu_id=asu_id, 
                wavelength=parser.wavelength,
            )
            if parser.separate:
                asu_id += 1
            if cell is None:
                cell = loader.cell
            _data = loader.get_dataset(
                num_cpus=parser.num_cpus,
                logging_level=parser.ray_log_level,
            )
            if data is None:
                data = _data
            else:
                data = data.concatenate(_data)

    elif all([_is_dials_file(f) for f in parser.inputs]):
        from abismal.io import StillsLoader
        expt_files = [f for f in parser.inputs if _is_expt_file(f)]
        refl_files = [f for f in parser.inputs if _is_refl_file(f)]

        data = None
        if parser.separate:
            for expt,refl in zip(expt_files, refl_files):
                loader = StillsLoader([expt], [refl], space_group, cell, parser.dmin, asu_id)
                asu_id += 1
                _data = loader.get_dataset()
                if data is None:
                    data = _data
                else:
                    data = data.concatenate(_data)
        else:
            loader = StillsLoader(
                expt_files, refl_files, space_group, cell, parser.dmin, asu_id=asu_id
            )
            data = loader.get_dataset()
            asu_id += 1
        if cell is None:
            cell = loader.cell
    else:
        raise ValueError(
            "Couldn't determine input file type. "
            "DIALS reflection tables and CrystFEL streams are supported."
        )
    if space_group is None:
        if hasattr(loader, 'spacegroup'):
            space_group = loader.spacegroup
        else:
            space_group = 'P1'

    # Gemmification
    if not isinstance(cell, gemmi.UnitCell):
        cell = gemmi.UnitCell(*cell)
    if not isinstance(space_group, gemmi.SpaceGroup):
        space_group = gemmi.SpaceGroup(space_group)

    # Handle setting up the test fraction, shuffle buffer, batching, etc
    test = None
    if parser.test_fraction > 0.:
        train,test = split_dataset_train_test(data, parser.test_fraction)
        test  = test.cache().repeat().ragged_batch(parser.batch_size)
        test = test.prefetch(AUTOTUNE)
    else:
        train = data
    train = train.cache().repeat()
    if parser.shuffle_buffer_size > 0:
        train = train.shuffle(parser.shuffle_buffer_size)
    train = train.ragged_batch(parser.batch_size)

    rasu = []
    for i in range(asu_id):
        rasu.append(ReciprocalASU(
            cell,
            space_group,
            parser.dmin,
            anomalous=parser.anomalous,
        ))
    rac = ReciprocalASUCollection(*rasu)

    reindexing_ops = ['x,y,z']
    if not parser.disable_index_disambiguation:
        ops = gemmi.find_twin_laws(
            cell,
            space_group,
            3.0,
            False
        )
        reindexing_ops = reindexing_ops + [op.triplet() for op in ops] 

    if parser.intensity_posterior:
        surrogate_posterior = GammaPosterior(
            rac, 
            kl_weight=parser.kl_weight,
            epsilon=parser.epsilon,
            scale_factor=parser.init_scale,
        )
    else:
        prior = None
        if parser.parents is not None:
            from abismal.surrogate_posterior.structure_factor.wilson import MultiWilsonPrior
            prior = MultiWilsonPrior(
                rac, 
                parser.parents, 
                parser.prior_correlation, 
                parser.reindexing_ops,
            )
        surrogate_posterior = FoldedNormalPosterior(
            rac, 
            kl_weight=parser.kl_weight,
            epsilon=parser.epsilon,
            scale_factor=parser.init_scale,
            prior=prior
        )

    scale_model = ImageScaler(
        mlp_width=parser.d_model, 
        mlp_depth=parser.layers, 
        hidden_units=parser.d_model * 2,
        activation=parser.activation,
        kl_weight=parser.scale_kl_weight,
        eps=parser.epsilon,
        num_image_samples=parser.sample_reflections_per_image,
    )

    if parser.studentt_dof is not None:
        from abismal.likelihood import StudentTLikelihood
        likelihood = StudentTLikelihood(parser.studentt_dof)
    else:
        from abismal.likelihood import NormalLikelihood
        likelihood = NormalLikelihood()

    model = VariationalMergingModel(
        scale_model, 
        surrogate_posterior, 
        likelihood=likelihood,
        mc_samples=parser.mc_samples,
        reindexing_ops=reindexing_ops,
    )

    if parser.learning_rate_final is not None:
        from tf_keras.optimizers.schedules import PolynomialDecay
        learning_rate = PolynomialDecay(
            parser.learning_rate,
            parser.epochs * parser.steps_per_epoch,
            end_learning_rate=parser.learning_rate_final,
        )
    else:
        learning_rate = parser.learning_rate

    opt = Adam(
        parser.learning_rate, 
        parser.beta_1, 
        parser.beta_2, 
        global_clipnorm=parser.global_clipnorm, 
        clipnorm=parser.clipnorm, 
        clipvalue=parser.clip, 
        epsilon=parser.adam_epsilon, 
        amsgrad=parser.amsgrad
    )

    
    mtz_saver = MtzSaver(parser.out_dir, parser.anomalous)
    history_saver = HistorySaver(parser.out_dir, gpu_id=parser.gpu_id)
    weight_saver  = ModelCheckpoint(
        filepath=f'{parser.out_dir}/abismal.weights.h5', save_weights_only=True, verbose=1)

    callbacks = [
        mtz_saver,
        history_saver,
        weight_saver,
    ]

    if parser.eff_files is not None:
        for i,eff_file in enumerate(parser.eff_files.split(',')):
            pfx = f"eff_{i}"
            if parser.anomalous:
                f = AnomalousPeakFinder(
                    parser.out_dir, eff_file, epoch_stride=parser.phenix_frequency, 
                    asu_id=0, output_prefix=pfx
                )
            else:
                f = PhenixRunner(
                    parser.out_dir, eff_file, epoch_stride=parser.phenix_frequency, 
                    asu_id=0, output_prefix=pfx
                )
            callbacks.append(f)

    train = train.prefetch(AUTOTUNE)

    model.compile(opt, run_eagerly=parser.run_eagerly)
    history = model.fit(
        x=train, 
        epochs=parser.epochs, 
        steps_per_epoch=parser.steps_per_epoch, 
        validation_steps=parser.validation_steps, 
        callbacks=callbacks, 
        validation_data=test
    )

if __name__=='__main__':
    main()




import tensorflow as tf
import tf_keras as tfk

"""
Various models that form convex combinations of inputs.
"""

class ConvexCombination(tfk.layers.Layer):
    def __init__(self, kernel_initializer='glorot_normal', dropout=None):
        super().__init__()
        if dropout is not None:
            self.dropout = tfk.layers.Dropout(dropout)
        else:
            self.dropout = None
        self.softmax = tfk.layers.Softmax(axis=-1)
        self.linear = tfk.layers.Dense(1, kernel_initializer=kernel_initializer, use_bias=False)

    def call(self, data, mask=None, encoding=None, **kwargs):
        if encoding is None:
            encoding = data
        score = self.linear(data)
        if mask is None:
            mask = tf.ones_like(score)
        if self.dropout is not None:
            mask = self.dropout(mask)
        score = tf.squeeze(score, axis=-1)
        score = self.softmax(score, mask=tf.squeeze(mask, axis=-1))
        encoded = tf.reduce_sum(score[...,None] * encoding, axis=-2, keepdims=True)
        return encoded

class ConvexCombinations(tfk.layers.Layer):
    def __init__(self, npoints, kernel_initializer='glorot_normal', dropout=None):
        super().__init__()
        if dropout is not None:
            self.dropout = tfk.layers.Dropout(dropout)
        else:
            self.dropout = None
        self.softmax = tfk.layers.Softmax(axis=-1)
        self.einsum = None
        self.npoints = npoints
        self.kernel_initializer = kernel_initializer

    def build(self, shape, **kwargs):
        self.einsum = tfk.layers.EinsumDense("...a,bac->...bc", kernel_initializer=self.kernel_initializer, output_shape=(self.npoints, 1), bias_axes='c')

    def call(self, data, mask=None, **kwargs):
        score = self.einsum(data)[...,0]
        if mask is None:
            mask = tf.ones_like(score)
        else:
            #TODO: is this correct??
            mask = np.ones_like(score) * mask[...,None,:]

        if self.dropout is not None:
            mask = self.dropout(mask)
        score = self.softmax(score, mask=mask)
        points = tf.einsum("...ab,...ac->...bc", score, data)
        return points

class Average(tfk.layers.Layer):
    def __init__(self, axis, keepdims=True):
        super().__init__()
        self.axis = axis
        self.keepdims = keepdims

    def call(self, data, **kwargs):
        return tf.reduce_mean(data, self.axis, self.keepdims)



import tensorflow as tf
import tf_keras as tfk
from tensorflow_probability import distributions as tfd
import numpy as np
from reciprocalspaceship.utils import hkl_to_asu,is_absent
from inspect import signature
from reciprocalspaceship.decorators import spacegroupify,cellify
import gemmi

class ReciprocalASUCollection(tfk.layers.Layer):
    def __init__(self, *reciprocal_asus):
        super().__init__()
        self.reciprocal_asus = reciprocal_asus
        asu_ids,dHKL,centric,epsilon = [],[],[],[]
        Hunique = []
        self.asu_size = 0
        self.hmax = 0
        for asu_id, rasu in enumerate(reciprocal_asus):
            dHKL.append(rasu.dHKL)
            centric.append(rasu.centric)
            epsilon.append(rasu.epsilon)
            self.asu_size = self.asu_size + rasu.asu_size
            self.hmax = np.maximum(self.hmax, rasu.hmax)
            asu_ids.append(tf.ones(rasu.asu_size, dtype='int32') * asu_id)
            Hunique.append(rasu.Hunique)

        self.centric = tf.concat(centric, axis=-1)
        self.epsilon = tf.concat(epsilon, axis=-1)
        self.dHKL = tf.concat(dHKL, axis=-1)
        self.asu_id = tf.concat(asu_ids, axis=-1)
        self.Hunique = tf.concat(Hunique, axis=-2)

        h,k,l = self.hmax
        self.miller_id = -np.ones((len(self), 2*h+1, 2*k+1, 2*l+1), dtype=np.int32)
        self.grid_size = tf.convert_to_tensor((2*h+1, 2*k+1, 2*l+1), dtype=tf.int32)
        offset = 0
        for asu_id, rasu in enumerate(reciprocal_asus):
            h,k,l = rasu.Hall.T
            self.miller_id[asu_id *np.ones_like(h),h,k,l] = rasu.miller_id[h,k,l] + offset
            offset = offset + rasu.asu_size

    def _ensure_in_range(self, H):
        """ Cast any out of bounds indices to [0,0,0] """
        idx = tf.reduce_any(tf.abs(H) > self.hmax, axis=-1, keepdims=True)
        out = tf.where(
            idx,
            tf.zeros_like(H),
            H
        )
        return out

    def _wrap_negative_millers(self, H):
        """ Wrap negative miller indices to positive for gather """
        return tf.where(H >= 0, H, H + self.grid_size[...,:])

    def _sanitize_millers(self, H):
        H = self._ensure_in_range(H)
        H = self._wrap_negative_millers(H)
        return H

    def _miller_ids(self, asu_id, H):
        H = tf.cast(H, 'int32')
        asu_id = tf.cast(asu_id, 'int32')
        hplus = self._sanitize_millers(H)
        ahkl = tf.concat((asu_id, hplus), axis=-1)
        return tf.gather_nd(self.miller_id, ahkl)

    def _gather(self, tensor, asu_id, H):
        idx = self._miller_ids(asu_id, H)
        safe = tf.maximum(idx, 0)
        return tf.gather(tensor, safe)

    def gather(self, tensor, asu_id, H):
        """
        Parameters
        ----------
        tensor : tf.Tensor 
            A tensor in the length of self.asu_size
        asu_id : tf.Tensor
            A potentially ragged tensor of asu ids
        H : tf.Tensor
            A potentially ragged tensor of miller indices
        """
        return tf.ragged.map_flat_values(self._gather, tensor, asu_id, H)

    def __iter__(self):
        return self.reciprocal_asus.__iter__()

    def __next__(self):
        return self.reciprocal_asus.__next__()

    def __len__(self):
        return len(self.reciprocal_asus)

class ReciprocalASU(tfk.layers.Layer):
    @cellify
    @spacegroupify
    def __init__(self, cell, spacegroup, dmin, anomalous=True):
        """
        Base Layer that remaps observed miller indices to the reciprocal asu.
        This class enables indexing of per reflection variables from tf.
        All the attributes are in the length of the number of non-redudant 
        miller indices.

        Attributes
        ----------
        asu_size : int
            Number of non-redundant reflections in the ASU
        dHKL : np.ndarray(np.float32)
            Reflection resolution
        centric : np.ndarray(bool)
            Indicator for centric vs acentric reflections
        epsilon : np.ndarray(np.float32)
            Reflection multiplicity. 

        Parameters
        ----------
        cell : gemmi.UnitCell
            A gemmi.UnitCell object
        spacegroup : gemmi.SpaceGroup
            A gemmi SpaceGroup object
        dmin : float
            The maximum resolution in Ångstroms
        anomalous : bool
            If true, treat Friedel mates as non-redudant
        """
        super().__init__()
        self.anomalous = anomalous
        self._cell = cell.parameters
        self._spacegroup = spacegroup.xhm()
        self.dmin = dmin
        go = self.spacegroup.operations()

        H = self.get_reciprocal_cell()

        Hasu,Isym = hkl_to_asu(H, spacegroup)
        if anomalous:
            friedel_sign = np.array([-1, 1])[Isym % 2][:,None]
            centric = go.centric_flag_array(Hasu).astype(bool)
            friedel_sign[centric] = 1
            Hasu = friedel_sign*Hasu

        self.Hunique,inv = np.unique(Hasu, axis=0, return_inverse=True)
        self.asu_size = len(self.Hunique)

        #This 3d grid contains the unique miller id for each reflection
        #-1 means that you shouldn't observe this miller index in the
        #data set
        h,k,l = self.hmax
        self.miller_id = -np.ones((2*h+1, 2*k+1, 2*l+1), dtype=np.int32)
        self.miller_id[H[:,0], H[:,1], H[:,2]] = np.arange(self.asu_size)[inv]
        self.Hall = H

        self.dHKL = cell.calculate_d_array(self.Hunique).astype(np.float32)
        self.epsilon = go.epsilon_factor_array(self.Hunique).astype(np.float32)
        self.centric = go.centric_flag_array(self.Hunique).astype(bool)

    @property
    def spacegroup(self):
        return gemmi.SpaceGroup(self._spacegroup)

    @property
    def cell(self):
        return gemmi.UnitCell(*self._cell)

    def get_reciprocal_cell(self):
        """ Generate the full reciprocal cell respecting systematic absences """
        go = self.spacegroup.operations()
        self.hmax = np.array(self.cell.get_hkl_limits(self.dmin))

        h,k,l = self.hmax

        self.grid_size = tf.convert_to_tensor((2*h+1, 2*k+1, 2*l+1), dtype=tf.int32)
        H = np.mgrid[-h:h+1:1,-k:k+1:1,-l:l+1:1].reshape((3, -1)).T

        #Apply resolution cutoff and remove absences
        d = self.cell.calculate_d_array(H)
        H = H[d >= self.dmin]
        H = H[~is_absent(H, self.spacegroup)]

        #Remove 0, 0, 0
        H = H[np.any(H != 0, axis=1)]

        return H

    def _ensure_in_range(self, H):
        """ Cast any out of bounds indices to [0,0,0] """
        idx = tf.reduce_any(tf.abs(H) > self.hmax, axis=-1, keepdims=True)
        out = tf.where(
            idx,
            tf.zeros_like(H),
            H
        )
        return out

    def _wrap_negative_millers(self, H):
        """ Wrap negative miller indices to positive for gather """
        return tf.where(H >= 0, H, H + self.grid_size[...,:])

    def _sanitize_millers(self, H):
        H = self._ensure_in_range(H)
        H = self._wrap_negative_millers(H)
        return H

    def _miller_ids(self, H):
        H = tf.cast(H, 'int32')
        hplus = self._sanitize_millers(H)
        return tf.gather_nd(self.miller_id, hplus)

    def gather(self, tensor, H):
        """
        Parameters
        ----------
        tensor : tf.Tensor 
            A tensor in the length of self.asu_size
        H : tf.Tensor
            A potentially ragged tensor of miller indices
        """
        idx = self._miller_ids(H)
        safe = tf.maximum(idx, 0)
        return tf.gather(tensor, safe)




import tensorflow as tf
import tf_keras as tfk
import numpy as np
import pandas as pd
from os.path import exists,dirname,abspath
from os import mkdir
from time import time
from subprocess import run
from shutil import which



class HistorySaver(tfk.callbacks.Callback):
    def __init__(self, output_directory, gpu_id=None,  *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.gpu_id=0
        self.output_directory = abspath(output_directory)
        self.start = time()

        if not exists(self.output_directory):
            mkdir(output_directory)
        self.records = []

    def on_epoch_begin(self, epoch, logs=None):
        self.save_history(epoch)

    def on_train_end(self, epoch, logs=None):
        epoch = self.records[-1]['Epoch'] + 1
        self.save_history(epoch)

    def save_history(self, epoch):
        out_file = f"{self.output_directory}/history.csv"

        record = {k:v[-1] for k,v in self.model.history.history.items()}
        record['Epoch'] = epoch
        record['Time (s)'] = time() - self.start

        if self.gpu_id is not None:
            for k,v in self.get_gpu_memory_usage().items():
                record[k] = v
        self.records.append(record)

        df = pd.DataFrame.from_records(self.records)
        df.to_csv(out_file, index=False)

    def get_gpu_memory_usage(self):
        usage = {}

        smi_path = which('nvidia-smi')
        if smi_path is None:
            return usage

        nvidia_smi_cmd = f"nvidia-smi -q -i {self.gpu_id} -d MEMORY"
        result = run(nvidia_smi_cmd.split(), capture_output=True)
        txt = result.stdout.decode()

        bank = None
        for line in txt.split('\n'):
            if 'Memory Usage'  in line:
                bank = ' '.join(line.split()[0:-2])
                continue
            if bank is None:
                continue
            if line=='':
                continue
            stat = line.split()[0]
            unit = line.split()[-1]
            k = f'{bank} {stat} ({unit})'
            v = int(line.split()[-2])
            usage[k] = v

        return usage





import tensorflow as tf


def split_dataset_train_test(data, test_frac, seed=1234):
    """ Deterministically split data into fractions """
    train = data.enumerate().filter(
        lambda i,x: tf.random.stateless_uniform([1], (i+seed, i+seed))[0] > test_frac
    ).map(
        lambda i,x: x
    )
    test  = data.enumerate().filter(
        lambda i,x: tf.random.stateless_uniform([1], (i+seed, i+seed))[0] <= test_frac
    ).map(
        lambda i,x: x
    )
    return train, test





def list_of_ints(string):
    return [int(i) for i in string.split(',')]

def list_of_floats(string):
    return [float(i) for i in string.split(',')]

def list_of_ops(string):
    return string.split(';')



from setuptools import setup, find_packages

# Get version number
def getVersionNumber():
    with open("abismal/VERSION", "r") as vfile:
        version = vfile.read().strip()
    return version

__version__ = getVersionNumber()

PROJECT_URLS = {
    "Bug Tracker": "https://github.com/rs-station/abismal/issues",
    "Source Code": "https://github.com/rs-station/abismal",
}


LONG_DESCRIPTION = """
Stochastic merging for diffraction data.
"""

setup(
    name='abismal',
    version=__version__,
    author='Kevin M. Dalton',
    author_email='kevinmdalton@gmail.com',
    license="MIT",
    include_package_data=True,
    packages=find_packages(),
    long_description=LONG_DESCRIPTION,
    description='Stochastic merging for diffraction data.',
    project_urls=PROJECT_URLS,
    python_requires=">=3.7,<3.12",
    url="https://github.com/rs-station/abismal",
    install_requires=[
        "reciprocalspaceship>=0.9.16",
        "tqdm",
        "tensorflow",
        "tf_keras",
        "tensorflow-probability",
        "rs-booster",
    ],
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'pytest-cov', 'pytest-xdist'],
    entry_points={
        "console_scripts": [
            "abismal=abismal.command_line.abismal:main",
        ]
    },
)



import numpy as np
import reciprocalspaceship as rs
import tensorflow as tf
import tensorflow_probability as tfp
import gemmi
from abismal.layers.standardization import Standardize
from tensorflow_probability import distributions as tfd
from tensorflow_probability import layers  as tfl
from tensorflow_probability import util as tfu
from tensorflow_probability import bijectors as tfb
from abismal.symmetry import Op
import tf_keras as tfk
from IPython import embed


class VariationalMergingModel(tfk.models.Model):
    def __init__(self, scale_model, surrogate_posterior, likelihood, mc_samples=1, eps=1e-6, reindexing_ops=None):
        super().__init__()
        self.eps = eps
        self.likelihood = likelihood
        self.scale_model = scale_model
        self.surrogate_posterior = surrogate_posterior
        self.mc_samples = mc_samples
        if reindexing_ops is None:
            reindexing_ops = ["x,y,z"]
        self.reindexing_ops = [Op(op) for op in reindexing_ops]
        self.standardize_intensity = Standardize(center=False)
        self.standardize_metadata = Standardize()

    def call(self, inputs, mc_samples=None, **kwargs):
        if mc_samples is None:
            mc_samples = self.mc_samples

        (
            asu_id,
            hkl,
            resolution,
            wavelength,
            metadata,
            iobs,
            sigiobs,
        ) = inputs
        if self.standardize_intensity is not None:
            iobs = tf.ragged.map_flat_values(self.standardize_intensity, iobs)
            sigiobs = tf.ragged.map_flat_values(self.standardize_intensity.standardize, sigiobs)
        if self.standardize_metadata is not None:
            metadata = tf.ragged.map_flat_values(self.standardize_metadata, metadata)

        training = kwargs.get('training', None)
        ll = None
        ipred = None

        q = self.surrogate_posterior.flat_distribution()
        ipred = q.sample(mc_samples)

        kl_div = self.surrogate_posterior.register_kl(ipred, asu_id, hkl, training)

        if self.surrogate_posterior.parameterization == 'structure_factor':
            ipred = tf.square(ipred)

        if training:
            self.surrogate_posterior.register_seen(asu_id, hkl)

        ipred = tf.transpose(ipred)
        ipred_scaled = None

        _inputs = (
            asu_id,
            hkl,
            resolution,
            wavelength,
            metadata,
            iobs,
            sigiobs,
        ) 
        scale = self.scale_model(
            _inputs, 
            mc_samples=mc_samples, 
            **kwargs
        )

        for op in self.reindexing_ops:
            # Choose the best indexing solution for each image
            _hkl = tf.ragged.map_flat_values(op, hkl)

            _ipred = self.surrogate_posterior.rac.gather(ipred, asu_id, _hkl)
            _ipred = _ipred * scale

            _ll = tf.ragged.map_flat_values(self.likelihood, _ipred, iobs, sigiobs)
            _ll = tf.reduce_mean(_ll, [-1, -2], keepdims=True)

            if ll is None:
                ipred_scaled = _ipred
                ll = _ll
            else:
                idx =  _ll > ll
                ipred_scaled = tf.where(idx, _ipred, ipred_scaled)
                ll = tf.where(idx, _ll, ll)

        self.likelihood.register_metrics(
            ipred_scaled.flat_values, 
            iobs.flat_values, 
            sigiobs.flat_values,
        )

        # This is the mean across mc samples and observations
        ll = tf.reduce_mean(ll) 

        self.add_metric(-ll, name='NLL')
        self.add_loss(-ll)

        ipred_avg = tf.reduce_mean(ipred_scaled, axis=-1)
        return ipred_avg

    #For production with super nan avoiding powers
    @tf.function
    def train_step(self, data):
        # Unpack the data. Its structure depends on your model and
        # on what you pass to `fit()`.
        x, y = data

        # Set up metrics dict
        metrics = {m.name: m.result() for m in self.metrics}

        with tf.GradientTape(persistent=True) as tape:
            y_pred = self(x, training=True)  # Forward pass
            # Compute the loss value
            # (the loss function is configured in `compile()`)
            loss = self.compiled_loss(y, y_pred, regularization_losses=self.losses)

        # Compute gradients
        scale_vars = self.scale_model.trainable_variables
        grad_scale = tape.gradient(loss, scale_vars)
        grad_s_norm = tf.sqrt(
            tf.reduce_mean([tf.reduce_mean(tf.square(g)) for g in grad_scale])
        )
        metrics["|∇s|"] = grad_s_norm

        q_vars = self.surrogate_posterior.trainable_variables
        grad_q= tape.gradient(loss, q_vars)
        grad_q_norm = tf.sqrt(
            tf.reduce_mean([tf.reduce_mean(tf.square(g)) for g in grad_q])
        )
        metrics["|∇q|"] = grad_q_norm

        trainable_vars = scale_vars + q_vars 

        gradients = grad_scale + grad_q 

        ll_vars = self.likelihood.trainable_variables
        if len(ll_vars) > 0:
            grad_ll = tape.gradient(loss, ll_vars)
            grad_ll_norm = tf.sqrt(
                tf.reduce_mean([tf.reduce_mean(tf.square(g)) for g in grad_ll])
            )
            trainable_vars += ll_vars
            gradients += grad_ll
            metrics["|∇ll|"] = grad_ll_norm

        gradients = [tf.where(tf.math.is_finite(g), g, 0.) for g in gradients]
        self.optimizer.apply_gradients(zip(gradients, trainable_vars))

        # Update metrics (includes the metric that tracks the loss)
        self.compiled_metrics.update_state(y, y_pred)


        # Return a dict mapping metric names to current value
        return metrics





import numpy as np
import tensorflow as tf
import tf_keras as tfk
from tensorflow_probability import stats as tfs


class RunningMoments:
    def __init__(self, axis=-2):
        self.n=0
        self.s=0
        self.mean=0
        self.axis=axis

    def update(self, x):
        k = len(x)
        self.n += k
        diff = x - self.mean
        self.mean = self.mean + np.sum(diff / self.n, axis=self.axis, keepdims=True)
        diff *= (x - self.mean)
        self.s = self.s + diff.sum(axis=self.axis, keepdims=True)

    @property
    def var(self):
        if self.n <= 1:
            return None
        return self.s / self.n

    @property
    def std(self):
        return np.sqrt(self.var)

class Standardize(tfk.layers.Layer):
    def __init__(self, center=True, max_counts=np.inf, epsilon=1e-6):
        super().__init__()
        self.epsilon = epsilon
        self.center = center
        self.max_counts = max_counts

    def build(self, shape):
        mom_shape = [1] * len(shape)
        mom_shape[-1] = shape[-1]
        reduce_dims = list(range(len(shape)-1))
        self.reduce_dims = reduce_dims
        self.axis_size = shape[-1]

        self.mean = self.add_weight(
            shape=mom_shape,
            initializer='zeros',
            dtype=tf.float32,
            trainable=False,
            name='mean',
        )
        self.m2= self.add_weight(
            shape=mom_shape,
            initializer='zeros',
            dtype=tf.float32,
            trainable=False,
            name='mean',
        )
        self.count = self.add_weight(
            shape=(),
            initializer='zeros',
            dtype=tf.float32,
            trainable=False,
            name='count',
        )

    @property
    def count_float(self):
        return tf.cast(self.count, self.mean.dtype)

    @property
    def std(self):
        return tf.sqrt(self.var)

    @property
    def var(self):
        m2 = tf.clip_by_value(self.m2, self.epsilon, np.inf)
        return m2 / self.count_float

    def update(self, x):
        k = tf.reduce_sum(tf.ones_like(x)) / self.axis_size
        self.count.assign_add(k)
        diff = x - self.mean
        new_mean = self.mean + \
            tf.reduce_sum(diff, axis=self.reduce_dims, keepdims=True) / self.count
        self.mean.assign(new_mean)
        diff *= (x - self.mean)
        self.m2.assign(
            self.m2 + tf.reduce_sum(diff, axis=self.reduce_dims, keepdims=True)
        )

    def standardize(self, data):
        if self.center:
            return (data - self.mean) / self.std
        return data / self.std

    def call(self, data, training=True):
        if self.max_counts > self.max_counts:
             training = False
        if training:
            self.update(data)
        return self.standardize(data)



if __name__=="__main__":
    l = 10_000
    d = 5
    x = np.arange(l)[...,None] * np.ones((l, d))
    x = x.astype('float32')
    n = RunningMoments()
    s = Standardize()
    for batch in np.split(x, 10):
        n.update(batch)
        s(batch)
        assert np.allclose(n.mean, s.mean)
        assert np.allclose(n.var, s.var)
        assert np.allclose(n.std, s.std)




import tensorflow.compat.v2 as tf
import tensorflow_probability as tfp

import tf_keras as tfk


class WAdam(tfk.optimizers.Optimizer):
    def __init__(
        self,
        learning_rate=0.001,
        beta=0.9,
        epsilon=1e-12,
        weight_decay=None,
        clipnorm=None,
        clipvalue=None,
        global_clipnorm=None,
        use_ema=False,
        ema_momentum=0.99,
        ema_overwrite_frequency=None,
        jit_compile=True,
        name="WAdam",
        **kwargs
    ):
        super().__init__(
            name=name,
            weight_decay=weight_decay,
            clipnorm=clipnorm,
            clipvalue=clipvalue,
            global_clipnorm=global_clipnorm,
            use_ema=use_ema,
            ema_momentum=ema_momentum,
            ema_overwrite_frequency=ema_overwrite_frequency,
            jit_compile=jit_compile,
            **kwargs
        )
        self._learning_rate = self._build_learning_rate(learning_rate)
        self.beta = beta
        self.epsilon = epsilon

    def build(self, var_list):
        """Initialize optimizer variables.

        Adam optimizer has 2 types of variables: momentums and velocities.

        Args:
            var_list: list of model variables to build Adam variables on.
        """
        super().build(var_list)
        if hasattr(self, "_built") and self._built:
            return
        self._built = True
        self._momentums = []
        self._velocities = []
        for var in var_list:
            self._momentums.append(
                self.add_variable_from_reference(
                    model_variable=var, variable_name="m"
                )
            )
            self._velocities.append(
                self.add_variable_from_reference(
                    model_variable=var, variable_name="v"
                )
            )

    def update_step(self, gradient, variable):
        """Update step given gradient and the associated model variable."""
        lr = tf.cast(self.learning_rate, variable.dtype)
        local_step = tf.cast(self.iterations + 1, variable.dtype)
        beta = tf.cast(self.beta, variable.dtype)

        var_key = self._var_key(variable)
        m = self._momentums[self._index_dict[var_key]]
        v = self._velocities[self._index_dict[var_key]]

        beta_power = tf.pow(tf.cast(self.beta, variable.dtype), local_step)
        alpha = lr / (1 - beta_power)

        if isinstance(gradient, tf.IndexedSlices):
            raise NotImplementedError("No sparse gradient implementation yet!")
            # Sparse gradients.
            m.assign_add(-m * (1 - self.beta_1))
            m.scatter_add(
                tf.IndexedSlices(
                    gradient.values * (1 - self.beta_1), gradient.indices
                )
            )
            v.assign_add(-v * (1 - self.beta_2))
            v.scatter_add(
                tf.IndexedSlices(
                    tf.square(gradient.values) * (1 - self.beta_2),
                    gradient.indices,
                )
            )
            variable.assign_sub((m * alpha) / (tf.sqrt(v) + self.epsilon))
        else:
            # Dense gradients.
            g = gradient
            delta = g - m
            m.assign_add((beta - 1.) * m + (1. - beta) * delta)
            v.assign_add((beta - 1.) * v + (1. - beta) * delta * (g - m))
            #variable.assign_sub((m * alpha) / (tf.sqrt(v + m*m) + self.epsilon))
            variable.assign_sub((m * alpha) / (tf.sqrt(v) + self.epsilon))

    def get_config(self):
        config = super().get_config()

        config.update(
            {
                "learning_rate": self._serialize_hyperparameter(
                    self._learning_rate
                ),
                "beta": self.beta,
                "epsilon": self.epsilon,
            }
        )
        return config





from .loader import DataLoader
from .mtz import MTZLoader
from .dials import StillsLoader
from .stream import StreamLoader
from .utils import split_dataset_train_test
from .tf_settings import set_gpu



import reciprocalspaceship as rs
import tf_keras as tfk
import tensorflow as tf
from os.path import exists,dirname,abspath
from os import mkdir
from subprocess import Popen,DEVNULL
from abismal.callbacks import MtzSaver


class PhenixRunner(tfk.callbacks.Callback):
    """ Run PHENIX periodically on the output. """
    def __init__(self, output_directory, eff_file, 
            epoch_stride=5, asu_id=0, output_prefix='phenix', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.output_prefix = output_prefix
        self.asu_id = asu_id
        self.eff_file = eff_file
        self.epoch_stride = epoch_stride
        self.output_directory = abspath(output_directory)
        self.postprocessing

        if not exists(self.output_directory):
            mkdir(output_directory)

    def on_epoch_end(self, epoch, logs=None):
        if self.eff_file is not None and (epoch + 1) % self.epoch_stride == 0:
            self.run_phenix(epoch)

    def run_phenix(self, epoch):
        mtz_file = f"{self.output_directory}/asu_{self.asu_id}_epoch_{epoch+1}.mtz"

        phenix_dir = f"{self.output_directory}/{self.output_prefix}_asu_{self.asu_id}_epoch_{epoch+1}"
        if not exists(phenix_dir):
            mkdir(phenix_dir)
        command = [
            'phenix.refine',
            self.eff_file,
            mtz_file,
        ]

        stderr = phenix_dir + '/stderr.txt'
        stdout = phenix_dir + '/stdout.txt'
        with open(stderr, 'w') as e, open(stdout, 'w') as o:
            p = Popen(
                command, 
                cwd=phenix_dir,
                stderr=e,
                stdout=o,
            )




import numpy as np
import tensorflow as tf
import tf_keras as tfk

class ScaleRange(tfk.layers.Layer):
    def __init__(self, minimum=-1, maximum=1, axis=-1):
        super().__init__()
        self.axis = axis
        self.freeze = False
        self.range = maximum - minimum
        self.min = minimum

    def build(self, shape):
        param_shape = [1 for i in shape]
        param_shape[self.axis] = shape[self.axis]
        self.reduce_dims = list(range(len(shape)))
        del self.reduce_dims[self.axis]
        self.low = self.add_weight(
            shape=param_shape,
            initializer=tfk.initializers.Constant(np.inf),
            dtype=tf.float32,
            trainable=False,
            name='mean',
        )
        self.high = self.add_weight(
            shape=param_shape,
            initializer=tfk.initializers.Constant(-np.inf),
            dtype=tf.float32,
            trainable=False,
            name='mean',
        )

    def call(self, data, training=None):
        if training:
            data_min = tf.reduce_min(data, axis=self.reduce_dims, keepdims=True)
            data_max = tf.reduce_max(data, axis=self.reduce_dims, keepdims=True)
            self.low.assign(tf.minimum(data_min, self.low))
            self.high.assign(tf.maximum(data_max, self.high))

        out = self.range * (data - self.low) / (self.high - self.low) + self.min
        return out

class PositionalEncoding(tfk.layers.Layer):
    """ Positionally encode the last axis of an array. """
    def __init__(self, num_frequencies):
        super().__init__()
        self.scale = ScaleRange(axis=-1)
        self.num_frequencies = num_frequencies

    def call(self, data, training=None):
        p = self.scale(data, training=training)
        d = p.get_shape()[-1]
        p = tf.repeat(p, self.num_frequencies, axis=-1)

        L = tf.range(0, self.num_frequencies, dtype='float32')
        f = np.pi * 2 ** L
        f = tf.tile(f, (d,))
        fp = f * p

        out = tf.concat((
            tf.sin(fp),
            tf.cos(fp),
        ), axis=-1)
        return out


if __name__=='__main__':
    data = np.random.random((100, 37, 10))
    pe = PositionalEncoding(4)
    out = pe(data, training=True)
    print(out.shape)




from .skip import FeedForward
from .combinations import ConvexCombination,ConvexCombinations,Average
from .transformer import TransformerAverage
from .positional_encoding import ScaleRange,PositionalEncoding
from .standardization import Standardize
